#!/usr/bin/env python
# coding: utf-8

"""
"""


import numpy as np
import collections # for defaultdict
#from scipy import optimize # it was needed for spring solution
from .candybox import CandyBox
from scipy import stats # for sample_like()
from . import f_models # for sample_like()
from .spring import get_spring_solution # for new IS function, for ISSI in future
from .welford import Welford # for PushAndPull




def get_1DS_sample(f, x_sub):
    """1DS stands for 1D sampling.
    1DS is, actually, an importance sampling method
    that, actually, does not calculate IS weights
    as f_density / h_density, but
    (arbitrary, nonuniformly) divides interval to subintervals.
    By using CDF (SF, actually) transformes subintervals 
    to 0-1 uniform probability line.
    Then gets one node as middle point of every subinterval,
    weights therefore are just interval widths itself.
    No sampling imprecisions are introduced, 
    therefore no spring, no correction are needed. 
    Usage:
    Function takes: f, x
    f: univariate (1D) distribution (scipy-compatible)
    x_sub: coordinates (n+1) of subintervals.
    Returns: x, weights
    x: n coordinates
    weights: same as IS weights
    """
    # x: 1, 2, 3
    # SF: 0.1, 0.01, 0.001
    u_sub = f.sf(x_sub)
    weights = u_sub[:-1] - u_sub[1:]
    np.abs(weights, out=weights)
    u = (u_sub[:-1] + u_sub[1:])
    #č u musí být float
    #č takže dělení dvěma nemusí robiť potíže 
    np.divide(u, 2, out=u)
    x = f.isf(u)
    return x, weights

#č chcu udělat menší IS funkci s výrovnáním odhadů
#č Síce tohle už umí ISSI, 
#č třída ale neočekává celkovou pravděpodobnost odlišnou od nuly.
#č Taky jeji funkcionalita je hodně přebytečná pro Ghull.
#č Zkratka, chcu funkci, která by na vstupu brala:
#č 1. vahy IS (tj. funkce nebude řešít souřadnice a hustoty)
#č 2. roušku (omezíme se na nejčastější případ dvou vzajemně vyloučujících jevů) a
#č 3. celkovou pravděpodobnost
#č a posleze vracela výrovnané odhady

def get_IS_estimation(weights, mask, p_overall=1):
    """Function calculates Importance Sampling estimators
    of two mutually excludive events (typical for indicator or Heaviside step function). 
    It performs correction based on a priory known overall probability.
    It takes numpy array of IS weghts (PDF / weighting function density),
    numpy boolean array and a priory known overall probability.
    Returns corrected estimators for true_event and false_event"""
    
    #č vstupní kontroly
    if np.all(mask): #č všecko je pravda - není co řešit
        return p_overall, 0
    if not np.any(mask): #č není žádná pravda - není žádné překvapení
        return 0, p_overall
    
    Nsim = len(weights)
    # there are two events:
    true_masked = weights[mask]
    false_masked = weights[~mask]
    #č podle IS vzorečků
    #č když uživatel bude hodně snažit, tak
    #č v mat. očekaváních může zajistit nuly
    #č nechcu zatím řešit
    true_mean = np.sum(true_masked)/Nsim
    false_mean = np.sum(false_masked)/Nsim
    if (true_mean + false_mean) == p_overall:
        return true_mean, false_mean
    #č podle toho, co vidím ve vzorcích - mám pocit, že se
    #č nůlový rozptyl může vyskytnout pouze v případě Nsim == 1, což
    #č zde od uživatele neočakávám. Nebo spolu s nulovým průměrem,
    #č který zatím taky nechcu řešit.
    #č (Ten ISSI je teda hrozně перемудрёный, řeší bůhvíco)
    #č Zde to nepotřebujem, udělame všecko jednoduše.
    true_var = (np.sum(true_masked**2)/Nsim - true_mean**2) / Nsim
    false_var = (np.sum(false_masked**2)/Nsim - false_mean**2) / Nsim
    
    #č pošleme to pružině. Ta je obecnější, na větší počet jevů
    #č proto bere jako vstup numpy matice
    lenghts = np.array((true_mean, false_mean))
    flexibilities = np.array((true_var, false_var))
    #č vrací taky numpy
    ls = get_spring_solution(lenghts, flexibilities, L=p_overall)
    #č rozbalíme 
    #č (na rovinu, dělám to jen aby pak kód bylo možné jednoduše přečíst)
    true_corrected_mean, false_corrected_mean = ls
    
    return true_corrected_mean, false_corrected_mean #(*ls,)




# One does not simply get IS estimations...
#č Získat IS odhady není natolik snadné...
#č Ghull totiž integruje po částech
#č a mně příšlo, že nejsprávnější volbou
#č bude i odhady zpracovavat po částech.
#č Musel jsem tedy vytvořit třídu (Welford) pro
#č průběžný výpočet průměrů a rozptylů.
#č Teď na jeji bazí postavíme IS odhadovačku
class PushAndPull: #ё Тяни-Толкай
    """Class calculates corrected Importance Sampling estimators
    of two mutually excludive events (typical for indicator or Heaviside step function). 
    It performs correction based on a priory known overall probability.
    It takes numpy array of IS weghts (PDF / weighting function density),
    numpy boolean array and a priory known overall probability."""
    def __init__(self):
        self.true_wf = Welford()
        self.false_wf = Welford()
    
    def add(self, weights, mask):
        #č add_sparse() - právě pro IS,
        #č aby se uvnitř netrapilo nulama
        n = len(mask)
        self.true_wf.add_sparse(weights[mask], n)
        self.false_wf.add_sparse(weights[~mask], n)
       
    @property 
    def mean(self):
        return self.true_wf.mean, self.false_wf.mean
        
    @property 
    def var(self):
        #č otazkou je rozptyl čeho tady máme na mysli
        #č Já bych zde spíše očekával rozptyl průměru
        #č takže rozptyl vah IS vydělíme celkovým počtem měření
        assert self.true_wf.n == self.false_wf.n
        true_var = self.true_wf.s2 / self.true_wf.n
        false_var = self.false_wf.s2 / self.false_wf.n
        return true_var, false_var
    
    def correct_means(self, p_overall=1):
        """Method calculates Importance Sampling estimators
        of two mutually excludive events. 
        It performs correction based on a priory known overall probability.
        It takes as a optional parameter a priory known overall probability.
        Returns corrected estimators for true_event and false_event"""
        
        # there are two events:
        true_mean = self.true_wf.mean
        false_mean = self.false_wf.mean
        #č vstupní kontroly
        if (true_mean + false_mean) == p_overall:
            return true_mean, false_mean
        if (true_mean == 0) and (false_mean == 0):
            half = p_overall / 2
            return half, half #č půl na půl
        if true_mean == 0: 
            return 0, p_overall
        if false_mean == 0: 
            return p_overall, 0
        
        
        #č Nulové rozptyly neřeším - neumím představit, za jakých 
        #č podmínek by mohly ve skutečné aplikaci výskytnout 
        #č (variana s jediným vzorkem mně nezajimá)
        #č striktně vzato, musíme zde vydělit počtem měření
        #č ale vysledek se dělením na stejné číslo nezmění.
        #č Dokonce můžeme se spokojit i prostě součtemi kvadratů
        true_var = self.true_wf.S
        false_var = self.false_wf.S
        
        #č pošleme to pružině. Ta je obecnější, na větší počet jevů.
        #č Proto bere jako vstup numpy matice
        lenghts = np.array((true_mean, false_mean))
        flexibilities = np.array((true_var, false_var))
        #č vrací taky numpy
        ls = get_spring_solution(lenghts, flexibilities, L=p_overall)
        #č rozbalíme 
        #č (na rovinu, dělám to jen aby pak kód bylo možné jednoduše přečíst)
        true_corrected_mean, false_corrected_mean = ls
        
        return true_corrected_mean, false_corrected_mean #(*ls,)
    
    
    
    

#č ISS a ISSI by bylo vhod předělat na spring a 
#č vyhodit ty blbosti s půlením intervalů.
#č Nechám to na potom, po zakroku by bylo potřebné 
#č všecko dobře otestovat, teď na to nemám ani kapacitu, ani zajem
#č Goal&Ghull ISSI nepotřebuje.

# bisect function is taken from somewhere in internet. 
# https://www.quora.com/How-does-one-correctly-compare-two-floats-in-Python-to-test-if-they-are-equal
# Hope it is not big issue.
def bisect(f, target, low, high):
    low, high = float(low), float(high)
    while low != high:
        mid = (low + high) / 2
        f_mid = f(mid)
        if f_mid == target:
            return mid
        elif f_mid > target:
           high = mid if mid != high else low
        else:
           low = mid if mid != low else high
    return low
    

#ё нам позарез нужен ещё один, свой собственный словник 
#č ten, na rozdil od defaultdict'a, neuklada chybejicí složky do slovníku
class DefaultDict(dict):
    def __init__(self, default_value=None):
        self.default_value = default_value
        
    def __missing__(self, key):
        return self.default_value
    
    
#
#č deme na to, koťě!
#
    
    #č IS, (n-2)-krátá realizace, n>>2, n→∞
def IS(f, h_model, space_from_h='R', space_to_f='G',  Nsim=int(1e4)):
    """
    space_from_h
    space_to_f
    matching - v jakých souřadnicích robím merge a jaká PDF použiváme?
    """
    
    
    # zgenerujeme vzorky
    # nic zajimavýho
    h = h_model(Nsim)
    
    
    # tady musíme provést jeden trik
    # totež jako v IS_like - ve výsledku dycky dostaneme f_model
    to_sample = f.new_sample(getattr(h, space_from_h), space_to_f) # R-ko smerdžíme ako G-čko
    w = to_sample.pdf(space_to_f) / h.pdf(space_from_h) # snad je to správně
    
    # vahy máme
    # zabalme do boxu
    # zbytek už nejsou naši starosti
    return CandyBox(to_sample, w=w)
        


def get_norm_plan(nis, nvar, mean=0, std=1, design=None):
    """
    mean: [0.05, 2, 100500]
    std: [0.05, 2, 100500]
    
    design(nis, nvar) should return sampling plan in hypercube (U space)!
    """
    
    if design is None:
        sampling_plan_G = np.random.randn(nis, nvar)
    else: #оӵ я, ярам, гиперкуб. Кинлэсь меда Гаусс вӧлмэтын дизайн куром на?
        #č transformací z U prostoru dochází ke ztratě přesnosti 
        #č a omezenému na 8.2 poloměru v Gaussovském prostoru.
        #č Jenže jsem si uvědomil, že těžko na to narazím,
        #č projevilo by se to při počtu vzorků řadově 10**15,
        #č to bych musel tejden generovat vzorky
        sampling_plan_G = stats.norm.ppf(design(nis, nvar))
    
    
    #č pdf spočteme na původním Gaussovském designu
    # sample_pdf(sample / alpha) / np.prod(alpha)
    
    ## desired: pdf = np.prod(stats.norm.pdf(sampling_plan_G) / std, axis=1)
    pdf = stats.norm.pdf(sampling_plan_G)
    pdf = np.divide(pdf, std, out=pdf)
    pdf = np.prod(pdf, axis=1)
    
    
    #č a teď pustíme se do výpočtu souřadnic
    ## desired: sampling_plan_N = (sampling_plan_G * std) + mean
    sampling_plan_N = sampling_plan_G; del(sampling_plan_G)
    sampling_plan_N = np.multiply(sampling_plan_N, std, out=sampling_plan_N)
    sampling_plan_N = np.add(sampling_plan_N, mean, out=sampling_plan_N)
    
    return sampling_plan_N, pdf


def IS_norm(f, mean=0, std=1, sampling_space='G', nis=1000, design=None):
    """
    mean: [0.05, 2, 100500]
    std: [0.05, 2, 100500]
    
    design(nis, nvar) should return sampling plan in hypercube (U space)!
    """
    
    sampling_plan_N, pdf = get_norm_plan(nis, f.nvar, mean, std, design)
    
    #č tady musíme provést jeden trik
    #č totež jako v IS_like - ve výsledku dycky dostaneme f_model
    to_sample = f.new_sample(sampling_plan_N, sampling_space) #č naše N-ko smerdžíme ako G-čko
    w = to_sample.pdf(sampling_space) / pdf #č snad je to správně
    
    #č vahy máme
    #č zabalme do boxu
    #č zbytek už nejsou naši starosti
    return CandyBox(to_sample, w=w)



    # for simplex: d = nvar+2 
    # for cell: d = base_r**2
def IS_like(f_plan, sampling_space='G', weights=None, nis=1000, d=1, design=None):
    """
    takes sample and returns sampling plan with the same cov (in case d=1)
    covariance matrix we'll divide by d
    """
    
    plan = getattr(f_plan, sampling_space)
    
                
    S_bc = np.cov(plan, rowvar=False, bias=True, aweights=weights)
    barycenter = np.average(plan, axis=0, weights=weights)
    
    
    #č matika
    w, v = np.linalg.eig(S_bc)
    
    # use IS sampling density with center equal to the simplex's barycenter
    # set the minimum distance as the standard deviation of IS densisty
    #č u stats.norm zadáváme směrodatnou odchylku, to je asi správné
    sigmas = np.sqrt(w/d) 
    mean = 0
    h_plan_N, pdf_N = get_norm_plan(nis, f_plan.nvar, mean, sigmas, design)
    
    #ёӵ здесь уже так легко не отделаемся. Трансформовать кароно.
    h_plan_bc = (v @ h_plan_N.T).T
    h_plan_sing = h_plan_bc + barycenter
    
    
    
    #č sice to má nazev h_plan, ale nese rozdělení a hustoty v f-ku
    #
    #č jelikož výtvaříme nový vzorek z "čístých" souřadnic
    #č máme zaručeno, že h_plan vždy bude f_modelem (podle současného kódu CandyBox)
    h_plan = f_plan.new_sample(h_plan_sing, sampling_space) 
    ## desired: w = h_plan.pdf(sampling_space) / pdf_N # snad je to správně
    w = np.divide(h_plan.pdf(sampling_space), pdf_N, out=pdf_N); del(pdf_N)
    
    #č vahy máme
    #č zabalme do boxu
    #č zbytek už nejsou naši starosti
    return CandyBox(h_plan, w=w)






    # for simplex: d = nvar+2 
    # for cell: it's not so easy
def sample_alike(plan, weights=None, nis=1000, d=1, design=None):
    """
    takes sample and returns sampling plan with the same cov (in case d=1)
    covariance matrix we'll divide by d
    
    Returns: sampling_plan, h_pdfs
    """
    nsim, ndim = plan.shape
    
    #ё зрада отменяется!
    #č Tahle funkce se pokusí aspoň něco nejak navzorkovat,
    #č ale musí být aspoň-alespoň dva vzorky.
    #č nechvám zodpovědnost za to na volajícím kódě!
    assert nsim > 1
    
    S_bc = np.cov(plan, rowvar=False, bias=True, aweights=weights)
    barycenter = np.average(plan, axis=0, weights=weights)
    
    
    
    if nsim < ndim + 1:
        S2 = np.diag(S_bc)
        if not np.all(S2 > 0):
            std = np.sqrt(np.sum(S2) / ndim / d) 
        else:
            std = np.sqrt(S2 / d)
        return get_norm_plan(nis, ndim, mean=barycenter, \
                    std=std, design=design)
                
    
    #č matika
    w, v = np.linalg.eig(S_bc)
    
    if not np.all(w > 0):
        S2 = np.diag(S_bc)
        if not np.all(S2 > 0):
            std = np.sqrt(np.sum(w) / ndim / d) 
        else:
            std = np.sqrt(S2 / d)
        return get_norm_plan(nis, ndim, mean=barycenter, \
                    std=std, design=design)
    
    # use IS sampling density with center equal to the simplex's barycenter
    # set the minimum distance as the standard deviation of IS densisty
    #č u stats.norm zadáváme směrodatnou odchylku, to je asi správné
    sigmas = np.sqrt(w/d) 
    mean = 0
    h_plan_N, pdf_N = get_norm_plan(nis, ndim, mean, sigmas, design)
    
    #ёӵ здесь уже так легко не отделаемся. Трансформовать кароно.
    h_plan_bc = (v @ h_plan_N.T).T
    h_plan_sing = h_plan_bc + barycenter
    
    return h_plan_sing, pdf_N


            
#č tím "TrueIS" bylo myšleno abosolutně matematicky korektní 
#č započítávání sad vzorků, vybraných z libovolných (navzajem odlišných)
#č rozdělení IS. 
#č To je v literatuře známo jako "multiple importance sampling"
#č Tahle třída (na rozdil od ISSI) musí udržovat v paměti souřadnice všech 
#č předchozích sad vzorku a jejich rozdělení.
#č Tím se to líší od ISSI, která počíta ze sad pouze statistiky 
#č a ukladá pouze průměry a rozptyly.
#č S touhle třídou se hralo v době prvních hraček s rejection samplingem
#č a bylo to spíše takový proof-of-concept, u kterého jsem uviděl, že funguje,
#č ale žádné extensivní testování nikdy nebylo provedeno a třída nikdy nebyla
#č skutečně použita v kódu WellMet.
#č Tehdy problemem bylo to, že pro učely rejection samplingu by bylo třeba
#č ukladat nejen souřadnice a rozdělení, ale ještě navic i stavy 
#č přislušných triangulací, což bylo už i pro mně moc. 

#č Má nějakou alfu, která "not used", netuším, k čemu sloužila.
#č Má metody pro výpočet vah a pro výpočet odhadů s (bacha!) úpravenými vzorečky. 
class TrueIS:
    def __init__(self, f, IS_mode='G'):
        self.f_sample = f()
        self.series_data = []
        self.IS_mode = IS_mode
        self.events = []
        
    def add_IS_serie(self, sample, events, alpha=1):
        """
        alpha is for rejection sampling, not used
        """
        self.series_data.append((sample, events, alpha))
        self.events += events
        if self.IS_mode == 'R':
            # jestli máme to právé vzorkovácí rozdělení - tak nemáme čo robiť
            self.f_sample.add_sample(sample) # smerdží se to po R
            # w like weights
            #wt.w = to_sample.pdf_R / wt.h.pdf_R
        else: #IS_mode == 'G':
            # tady musíme provést jeden trik
            self.f_sample.add_sample(sample.R, 'G') # R-ko smerdžíme ako G-čko
            #wt.w = to_sample.pdf_G / wt.h.pdf_R # snad je to správně
        
        
    def get_weights(self):
        accu_denom_w = np.zeros(self.f_sample.nsim, dtype=float)
        if self.IS_mode == 'R':
            for serie in self.series_data:
                h_sample, events, alpha = serie
                h = h_sample.new_sample(self.f_sample)
                accu_denom_w += h.pdf_R * h_sample.nsim
            return self.f_sample.pdf_R / accu_denom_w, self.events
                
        else: # IS_mode == 'G'
            for serie in self.series_data:
                h_sample, events, alpha = serie
                h = h_sample.new_sample(self.f_sample.G, 'R')
                accu_denom_w += h.pdf_R * h_sample.nsim
            return self.f_sample.pdf_G / accu_denom_w, self.events
                
        
    def get_means(self):
        w, events = self.get_weights()
        keys = np.unique(events)
        Nsim = len(w)
        
        means = np.empty(len(keys), dtype=float)
        vars = np.empty(len(keys), dtype=float)
        
        
        # tak bacha
        ev = np.array(events)
        for key, i in zip(keys, range(len(keys))):
                
            mask = ev==key
            ev_w = w[mask]
            # podle IS vzorečků (skoro)
            key_mean = np.sum(ev_w)
            key_var = (np.sum(ev_w**2)*Nsim - key_mean**2) / Nsim
                
            # uložime
            means[i] = key_mean
            vars[i] = key_var
            
            # vyfiltrujeme zbytek
            w = w[~mask]
            ev = ev[~mask]
        
        # kontrola  
        assert len(w)==0 and len(ev)==0, "Что за хренотень?!"
        
        return means, vars
        

class ISSI:
    """
    IS statistics = weights series + events
    ISSI calculates probabilities of repeated IS series
    with implicit (non-zero) variances
    
    zda se mi, že tím nelze nic zhoršit
    """
    #č mám pocit, že je tu někde bug pythonu
    #č ten prazdnej list [] přežije cokoliv
    def __init__(self, events=None):
        """
        """ 
        self.series_data = []
        if events is None:
            self.events = []
        else:
            self.events = events
        
    def add_IS_serie(self, weights, events, implicit_multiplicator=1):
        keys = np.unique(events)
        Nsim = len(weights)
        
        # vytvoříme slovník pro sadu vzorků, který implicitně
        # bude předpokládát průměr=0 a rozptyl 1/Nsim^2 
        # pro všecko, co se v sadě neobjevilo
        if Nsim == 1:
            # jedno meření je taky měření)
            implicit_var = implicit_multiplicator
        else:
            implicit_var = implicit_multiplicator*(1-1/Nsim)/Nsim**2
        serie = DefaultDict((0, implicit_var))
        
        # tak bacha
        w = np.array(weights)
        ev = np.array(events)
        for key in keys:
            if key not in self.events:
                self.events.append(key)
                
            mask = ev==key
            ev_w = w[mask]
            # podle IS vzorečků
            key_mean = np.sum(ev_w)/Nsim
            key_var = (np.sum(ev_w**2)/Nsim - key_mean**2) / Nsim
            if key_var == 0: # to nechcem
                key_var = implicit_var
                
            # uložime
            serie[key] = (key_mean, key_var)
            
            # vyfiltrujeme zbytek
            w = w[~mask]
            ev = ev[~mask]
        
        # kontrola  
        assert len(w)==0 and len(ev)==0, "Что за хренотень?!"
        
        self.series_data.append(serie)
        
        # ачиз значение утёз
        self.get_estimations()
        
        
        
    def add_single_event_data(self, weights, event, nis=None, implicit_multiplicator=np.inf):
        
        if nis is None:
            Nsim = len(weights)
        else:
            Nsim = nis
            
        # kontrola předpokladů
        assert len(weights) <= Nsim, "jseš jistej, že máš spravnej počet měření?"
        
        # vytvoříme slovník pro sadu vzorků, který implicitně
        # bude předpokládát průměr=0 a rozptyl 1/Nsim^2 
        # pro všecko, co se v sadě neobjevilo
        if Nsim == 1:
            # jedno meření je taky měření)
            implicit_var = implicit_multiplicator
        else:
            implicit_var = implicit_multiplicator*(1-1/Nsim)/Nsim**2
        serie = DefaultDict((0, implicit_var))
        
        
        w = np.array(weights)
        if event not in self.events:
            self.events.append(event)
            
        # podle IS vzorečků
        key_mean = np.sum(w)/Nsim
        key_var = (np.sum(w**2)/Nsim - key_mean**2) / Nsim
        if key_var == 0: # to nechcem
            key_var = implicit_var
            
        # uložime
        serie[event] = (key_mean, key_var)
        
        self.series_data.append(serie)
        
        #č nebudeme hned počítat odhady, nechť to dělá volající kód sam
        # ачиз значение утёз
        #self.get_estimations()
        #self.estimations = {self.events[i] : self.values[i] for i in range(len(self.events))}
    
        
    def get_means(self):
        # počet jevů
        # number of events
        n_ev = len(self.events)
        self.weighted_means = np.empty(n_ev, dtype=float)
        self.weighted_vars = np.empty(n_ev, dtype=float)
        
        # spočteme važené průměry a rozptyly
        for event, key in zip(self.events, range(n_ev)):
            key_accusum = 0
            key_accuweight = 0
            for serie in self.series_data:
                key_mean, key_var = serie[event]
                key_accusum += key_mean / key_var
                key_accuweight += 1/key_var
            
            if key_accuweight == 0:
                #č já vidím, že .get_estimations nuly skryje
                self.weighted_means[key] = 0
                self.weighted_vars[key] = np.inf
            else:
                self.weighted_means[key] = key_accusum / key_accuweight
                self.weighted_vars[key] = 1/key_accuweight
            
        
        return self.weighted_means, self.weighted_vars, np.array(self.events)
            
    def get_estimations(self):
    
        # p-čka. We are still talking about probabilities, right?    
        p, vars, __ = self.get_means() 
        
        sum_p = np.sum(p)
        if sum_p == 1: # а вдруг?
            self.values = p
            #♥оӵ ачиз значение утёз
            self.estimations = {self.events[i] : self.values[i] for i in range(len(self.events))}
            return p, np.array(self.events)
            
            # spring (non-linear) analogue
        # silu, kterou zatížíme system jen tak neseženeme
        elif sum_p > 1: # stlačit
            low_atF = -np.pi/2
            high_atF = 0
        else: # roztahnout
            low_atF = 0
            high_atF = np.pi/2
        
        # nuly nám udělaj problém    
        mask = p==0
        p = np.ma.array(p)
        vars = np.ma.array(p)
        p.mask = mask
        vars.mask = mask
        
        F = np.tan(bisect(lambda atF: np.sum(np.exp(np.tan(atF)*vars/p)*p), 1, low_atF, high_atF))
        corrected_means = p * np.exp(F * vars/p)
        
        corrected_means.mask = False
        self.values = corrected_means
        #♥ ачиз значение утёз
        self.estimations = {self.events[i] : self.values[i] for i in range(len(self.events))}
        return corrected_means, np.array(self.events)
        
        
    def delete_event_data(self, event):
        """
        pomocná funkce třeba pro případ změny (zpřesnění) nějakých jevů na straně modelu
        function doesn't recalculate estimations after that!
        """
        # prozatím nebudu kontrolovat, zda ten jev tam vůbec je
        self.events.remove(event)
        for serie in self.series_data:
            # vymažem jev
            if event in serie:
                serie.pop(event)
        
        #č krucinál, tady narazím už ne natolik triviálné chyby
        for i in range(len(self.series_data)-1, -1, -1): # reverse
            serie = self.series_data[i]
            #č zkontrolujem, zda v serii se ještě něco zůstalo
            if len(serie) == 0:
                #č jako doufám, že tam tato serie bude jedina 
                self.series_data.remove(serie)
                
        
        
        
        
        


class ISS:
    """
    IS statistics = weights series + events
    ISS calculates probabilities of repeated IS series
    """
    def __init__(self):
        """
        defaultdict zajistí, že na každý jev nás čeka seznam, do kterého 
        může přidávat průměr a rozptyl ze sady
        """ 
        self.data = collections.defaultdict(list)
        
    def add_IS_serie(self, weights, events):
        keys = np.unique(events)
        Nsim = len(weights)
        
        # tak bacha
        w = np.array(weights)
        ev = np.array(events)
        for key in keys:
            mask = ev==key
            ev_w = w[mask]
            # podle IS vzorečků
            key_mean = np.sum(ev_w)/Nsim
            key_var = (np.sum(ev_w**2)/Nsim - key_mean**2) / Nsim
            # uložime
            self.data[key].append((key_mean, key_var))
            
            # vyfiltrujeme zbytek
            w = w[~mask]
            ev = ev[~mask]
        
        # kontrola  
        assert len(w)==0 and len(ev)==0, "Что за хренотень?!"
        
        
        
    def get_means(self):
        # počet jevů
        # number of events
        weighted_means = []
        weighted_vars = []
        
        # spočteme važené průměry a rozptyly
        for key_data in self.data.values():
            key_accusum = 0
            key_accuweight = 0
            for key_mean, key_var in key_data:
                key_accusum += key_mean / key_var
                key_accuweight += 1/key_var
            
            weighted_means.append(key_accusum / key_accuweight)
            weighted_vars.append(1/key_accuweight)
            
            
        return weighted_means, weighted_vars, np.array(list(self.data.keys()))
            
    def get_estimations(self):
        weighted_means, weighted_vars, __ = self.get_means() 
        
        # p-čka. We are still talking about probabilities, right?    
        p = np.array(weighted_means)
        vars = np.array(weighted_vars)
        
        sum_p = np.sum(p)
        if sum_p == 1: # а вдруг?
            return p, np.array(list(self.data.keys()))
            
            # spring (non-linear) analogue
        # silu, kterou zatížíme system jen tak neseženeme
        elif sum_p > 1: # stlačit
            low_atF = -np.pi/2
            high_atF = 0
        else: # roztahnout
            low_atF = 0
            high_atF = np.pi/2
            
        F = np.tan(bisect(lambda atF: np.sum(np.exp(np.tan(atF)*vars/p)*p), 1, low_atF, high_atF))
        corrected_means = p * np.exp(F * vars/p)
        
        return corrected_means, np.array(list(self.data.keys()))
        
        
        
            
           # scipy проверяет брэкеты, чем вызывает overflow warning, раздражает
#        sol = optimize.root_scalar(lambda atF: np.sum(np.exp(np.tan(atF)*vars/p)*p)-1, bracket=(-np.pi/2, np.pi/2) )
#        F = np.tan(sol.root)
        
        
        
        
        
                # triangle analogue        
#        if len(weighted_means) == 2:
#           # tak je to jednoduchý
#           # it's piece of cake
#           # triangle analogue
#           a, b = *weighted_means
#           # derivations of Heron's formula
#           da = 8*a*b**2 + 6*a - 2*b - 2
#           db = 8*b*a**2 + 6*b - 2*a - 2
#           # matice koeficientů přetvořené podmínkové rovnice
#           B = np.array([[1,1], [da, db]])

#           u = 1 - np.sum(p)
#           # vzoreček z teorii chyb měření a vyrovnávacího počtu nefunguje
#           # dává aj záporné pravděpodobnosti
#           #k = (1 - np.sum(weighted_means))/np.sum(weighted_vars)
#           #corrected_means = weighted_means + weighted_vars*k



import sys

from ..testcases import testcases_nD
from ..testcases import gaussian_2D, testcases_2D, testcases_2D_papers, testcases_nD_papers
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtWidgets

 

unparametrized_testcases = [
                            (gaussian_2D, "Gaussian 2D testcases"),
                            (testcases_2D_papers, "Well-known 2D testcases"),
                            (testcases_nD_papers, "Well-known nD testcases"),
                            (testcases_2D, "2D testcases"),
                            ]



class TestCasesListWidget(QtWidgets.QListWidget):
    def __init__(self, module, parent=None):
        self.module = module
        super().__init__(parent)
        self.addItems(module.__all__)
        
    def extract_selected_item(self):
        item = self.currentItem()
        if item is None:
            return getattr(self.module, self.module.__all__[0])
        else:
            item_str = self.currentItem().text()
            return getattr(self.module, item_str)


class SelectTestCaseWidget(pg.LayoutWidget):
    def __init__(self, parent=None):
        #č nejdřív vytvořiť apku
        self.app = pg.mkQApp("WellMet")
        # 
        super().__init__(parent)
        
        self.setup()
        
        #č zobraziť
        self.show()
        #č a spustit smyčku
        self.app.exec_()
        
    def setup(self):
        self.setWindowTitle("Select testcase")
        self.box = None
        
        self.setup_tabs()
        self.addWidget(self.tabs, row=0, col=0, rowspan=1, colspan=2)
        
        self.btn_exit = QtWidgets.QPushButton('Exit')
        self.addWidget(self.btn_exit, row=1, col=0)
        self.btn_exit.clicked.connect(self.exit)
        
        self.btn_choose = QtWidgets.QPushButton('Next')
        self.addWidget(self.btn_choose, row=1, col=1)
        self.btn_choose.clicked.connect(self.choose)
        
    def exit(self):
        self.app.quit()
        sys.exit()
        
    def choose(self):
        tab_index = self.tabs.currentIndex()
        if tab_index == 0:
            ndim, ok = QtWidgets.QInputDialog.getInt(self,\
                    "Enter dimension of the problem","ndim", value=2, min=2)
            if not ok:
                ndim = 2 
            testcase_helper = self.tab_nd.extract_selected_item()
            self.box = testcase_helper(ndim)
        else:
            tab_widget = self.tab_widgets[tab_index-1]
            testcase_helper = tab_widget.extract_selected_item()
            self.box = testcase_helper()
        self.app.quit()
        
    def setup_tabs(self):
        # Initialize tab screen
        self.tabs = QtWidgets.QTabWidget(self)
        
        ## Add tabs
        # nD - parametrized
        self.tab_nd = TestCasesListWidget(testcases_nD, self)
        self.tabs.addTab(self.tab_nd, "nD testcases")
        
        # the rest - unparametrized
        self.tab_widgets = []
        for testcases, description in unparametrized_testcases:
            tab_widget = TestCasesListWidget(testcases, self)
            self.tabs.addTab(tab_widget, description)
            #č ať nám pythonovej GC nekrešne appku
            self.tab_widgets.append(tab_widget)
        




#!/usr/bin/env python
# coding: utf-8

import pyqtgraph as pg
from pyqtgraph.Qt import QtGui, QtWidgets
from pyqtgraph.Qt import QtCore

import numpy as np
import pandas as pd # required for estimation graph



from .. import stm_df


#č nastavovaní mezí není filosofií pyqtgraph
#č v jejich examples byly použity pouze jednou - 
#č v ukázce možností ViewBox.
#č Už vím proč - sice zoomovaní příjemnější, když graf nikam neutiká,
#č ale až se naraží na limity, pg začně měnit proporce, 
#č které se už nezmění zpatky jen otačením kolečka myši.
#č Navíc setLimits nešikovně lame autoRange
def set_xlimits(plot_widget, xmin_lim, xmax_lim):
    #č není veřejná, není ale ani soukromá. vrací [x, y]
    x, y = plot_widget.getPlotItem().getViewBox().autoRangeEnabled()
    plot_widget.setLimits(xMin=xmin_lim, xMax=xmax_lim)
    if not x:
        [[xmin, __xmax], __y_range] = plot_widget.getPlotItem().viewRange()
        #č chcu aby x pokryval i čerstvě přidáná data
        plot_widget.setRange(xRange=(xmin, xmax_lim), padding=0)
        
    #č ten uvnitř volá i autoRange() na přislušných osech
    plot_widget.enableAutoRange(x=x, y=y)



def update_xrange(plot_widget, xmax):
    #č není veřejná, není ale ani soukromá. vrací [x, y]
    x, y = plot_widget.getPlotItem().getViewBox().autoRangeEnabled()
    if not x:
        [[xmin, _xmax], __y_range] = plot_widget.getPlotItem().viewRange()
        #č chcu aby x pokryval i čerstvě přidáná data
        plot_widget.setRange(xRange=(xmin, max(_xmax, xmax)), padding=0)
        
        #č ten uvnitř volá i autoRange() na přislušných osech
        plot_widget.enableAutoRange(x=x, y=y)







class GRaph(pg.PlotWidget):
    update_xrange = update_xrange
    
    def __init__(self, stream, box_data=None, parent=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        #č tím potokem je myšleno hlavní okínko
        self.stream = stream
        self.box_data = box_data
        if box_data is not None:
            box_data.estimations_updated.connect(self.stm_replot)
            self.setup_context_menu(box_data.TRI_menu)
        
        self.stream.box_runned.connect(self.replot)
        self.stream.slice_changed.connect(self.replot)
            
        self.stream.redraw_called.connect(self.redraw)
        
        self.redraw()
    
        
        
        
        
        
    def redraw(self):
        self.clear()
        self.setBackground('w')
        #č y limita v podstatě znemožní Log y graf
        #size=self.w.px_size*1.5
        
        # box part
        pos = () #np.empty((nsim, 4))
        size = self.stream.px_size * 2
        self.failures = self.plot(pos, pen=None, symbol='x', symbolPen='r',\
                    symbolSize=size*1.5,  name='Failures') # symbolBrush=0.2,
        self.failures.setZValue(100)
        
        self.proxy_failures = self.plot(pos, pen=None, symbol='p', symbolPen=0.5,\
                symbolSize=size, symbolBrush=(217,83,25), name='Proxy failures')
        self.proxy_failures.setZValue(95)
        
        self.successes = self.plot(pos, pen=None, symbol='+', \
                    symbolSize=size*1.5, symbolPen='g',  name='Successes')
        self.successes.setZValue(90)
        
        self.proxy_successes = self.plot(pos, pen=None, symbol='p', symbolPen=0.5, \
                symbolSize=size, symbolBrush=(119,172,48), name='Proxy successes')
        self.proxy_successes.setZValue(85)
        
        
        self.nodes = self.plot(pos, pen=None, symbol='o', symbolPen=0.5, \
                symbolSize=size, name='Nodes')
        self.nodes.setZValue(80)
        
        
        self.f_max_ruler = self.addLine(y=0, pen='r') 
        self.f_min_ruler = self.addLine(y=0, pen='r') 
            
        pen = pg.mkPen(color='b', width=1.5) # blue
        if hasattr(self.stream.sample_box, 'r_exact'):
            exact_name = "r_exact"
            y = self.stream.sample_box.r_exact
            self.r_exact = self.addLine(y=y, pen=pen, name=exact_name) 
            #č aby se nám něco zobrazovalo v legendu
            self.r_exact_PR = self.plot((), (), pen=pen, name=exact_name)
            self.f_min_ruler.show()
        
        if hasattr(self.stream.sample_box, 'R_exact'):
            exact_name = "R_exact"
            y = self.stream.sample_box.R_exact
            self.R_exact = self.addLine(y=y, pen=pen, name=exact_name) 
            #č aby se nám něco zobrazovalo v legendu
            self.R_exact_PR = self.plot((), (), pen=pen, name=exact_name)
            self.f_max_ruler.show()
        else:
            self.f_max_ruler.hide()
        
        self.replot()
        
        
        
        # estimation part
        if self.box_data is not None:
            x = y = () # zde jen vytvoříme kostru, nakrmime daty v .redraw()
            
            #xkcd_green = (167, 255, 181) # xkcd:light seafoam green #a7ffb5
            green = (159, 255, 173) 
            xkcd_red   = (253, 193, 197) # xkcd: pale rose (#fdc1c5)
            red = xkcd_red #(253, 0, 17)
            #xkcd_cream = (255, 243, 154) # let's try xkcd: dark cream (#fff39a)
            cream = (255, 242, 159)
            grey = (233, 233, 233)
            
            self.r_success = self.plot(x, y) 
            self.r_success.setZValue(-1000)
            self.R_success = self.plot(x, y) 
            self.R_success.setZValue(-1000)
            
            self.fill_s = pg.FillBetweenItem(self.r_success, self.R_success)
            #self.fill_mix.setData(name="mixed simplices measure")
            self.fill_s.setBrush(green)
            self.fill_s.setZValue(-1000)
            self.addItem(self.fill_s)
            
            self.r_f = self.plot(x, y)
            self.r_f.setZValue(-500)
            self.R_f = self.plot(x, y)
            self.R_f.setZValue(-500)
            
            self.fill_f = pg.FillBetweenItem(self.r_f, self.R_f)
            #self.fill_mix.setData(name="mixed simplices measure")
            self.fill_f.setBrush(red)
            self.fill_f.setZValue(-500)
            self.addItem(self.fill_f)
            
            self.r_mix = self.plot(x, y)
            self.r_mix.setZValue(-150)
            self.R_mix = self.plot(x, y)
            self.R_mix.setZValue(-150)
            
            self.fill_mix = pg.FillBetweenItem(self.r_mix, self.R_mix)
            #self.fill_mix.setData(name="mixed simplices measure")
            self.fill_mix.setBrush(cream)
            self.fill_mix.setZValue(-150)
            self.addItem(self.fill_mix)
            
            self.r = self.plot(x, y)
            self.r.setZValue(-10)
            self.R = self.plot(x, y)
            self.R.setZValue(-10)
            
            #self.pen_outside = self.plot(x, y)
            self.fill_outside = pg.FillBetweenItem(self.r, self.R)
            #self.fill_outside.setData(name="out of sampling domain estimation")
            self.fill_outside.setBrush(grey)
            self.fill_outside.setZValue(-10)
            self.addItem(self.fill_outside)
        
            self.stm_replot(self.box_data.df)
        
        
        
        
        
    def replot(self):
        nsim = self.stream.slider.value()
        
        sample_box = self.stream.sample_box[:nsim]
        
        lengths = np.sum(np.square(sample_box.G), axis=1)
        lengths = np.sqrt(lengths, out=lengths) #lengths of each radius-vector
        
        
        pos = np.empty((nsim, 2))
        pos[:,0] = np.arange(nsim)
        pos[:,1] = lengths
        
        if hasattr(sample_box, 'failsi'): #č to je normálně sample_box
            failsi = sample_box.failsi
            
            if np.any(failsi):
                f_radia = lengths[failsi]
                if hasattr(self.stream.sample_box, 'R_exact'):
                    self.f_max_ruler.setPos(np.max(f_radia))
                    self.f_max_ruler.show()
                self.f_min_ruler.setPos(np.min(f_radia))                
                self.f_min_ruler.show()
            else:
                self.f_max_ruler.hide()
                self.f_min_ruler.hide()
            
            try: # proxy denotes implicitly-known values
                proxy = sample_box.proxy.astype(bool)
            except AttributeError:
                proxy = np.full(nsim, False, dtype=bool)
            
            mask = np.all((failsi, ~proxy), axis=0)
            self.draw(self.failures, pos[mask])
            
            mask = np.all((~failsi, ~proxy), axis=0)
            self.draw(self.successes, pos[mask])
            
            mask = np.all((failsi, proxy), axis=0)
            self.draw(self.proxy_failures, pos[mask])
            
            mask = np.all((~failsi, proxy), axis=0)
            self.draw(self.proxy_successes, pos[mask])
            
        else: #č není to teda sample_box... 
            #č snad se nám povede nakreslit aspoň tečky?
            self.draw(self.nodes, pos)
            self.f_max_ruler.hide()
            self.f_min_ruler.hide()
        
            
    @staticmethod
    def draw(plot_item, data):
        #č musím to udělat takhle
        #č jinač to zlobí při posunutích slajderu
        if len(data):
            plot_item.setData(data)
            plot_item.show()
        else:
            plot_item.hide()





    def setup_context_menu(self, menu):
        # creates instance of LegendItem 
        # and saves it into plotItem.legend
        self.legend = self.addLegend()
    
        self.plotItem.ctrl.xGridCheck.setChecked(True)
        self.plotItem.ctrl.yGridCheck.setChecked(True)
        
        # delete build-in Transforms (with Log_x and Log_y) options, 
        # they can cause uncachable exception (on any zero in data) and crash
        self.plotItem.ctrlMenu.removeAction(self.plotItem.ctrlMenu.actions()[0])
        
        #č já se bojím. radší to uložím
        self.custom_menu = self.plotItem.vb.menu.addMenu("TRI options")
        
        self.plotItem.vb.menu.addMenu(menu)
        
        self.legend_chk = QtGui.QAction("Legend", self.custom_menu) 
        self.legend_chk.setCheckable(True)
        self.legend_chk.triggered.connect(lambda: self.legend.setVisible(self.legend_chk.isChecked()))
        self.custom_menu.addAction(self.legend_chk)
        # apply custom menu option
        self.legend.setVisible(self.legend_chk.isChecked())

        self.laction = QtGui.QAction("Show labels", self.custom_menu)
        self.laction.triggered.connect(self.show_labels)
        self.custom_menu.addAction(self.laction)
        
            
    def show_labels(self):
        self.setLabel('left', "Distsnce from origin in G space")
        self.setLabel('bottom', "Number of simulations")
    
        
    @staticmethod
    def _set_data(plot_item, serie):
        x, y = serie.index.to_numpy() - 1, serie.to_numpy()
        mask = y >= 0
        plot_item.setData(x[mask], y[mask])


    def stm_replot(self, df):
        if len(df) < 2: #č nevím proč hazí chyby. Asi kvůli zadané širce.
            return 
        
        if 'r' in df:
            self._set_data(self.r, df['r'])
            self._set_data(self.R, df['R'])
        
        if 'r_safe' in df:
            self._set_data(self.r_success, df['r_safe'])
            self._set_data(self.R_success, df['R_safe'])
            
        if 'r_mixed' in df:
            self._set_data(self.r_mix, df['r_mixed'])
            self._set_data(self.R_mix, df['R_mixed'])
            
            
        if 'r_failure' in df:
            self._set_data(self.r_f, df['r_failure'])
            self._set_data(self.R_f, df['R_failure'])
        
        
        xmax = df.index.max() 
        self.update_xrange(xmax*1.02)








"""
=============
График виӝет 
Grafy
Estimation graph widgets
========================
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
"""
        
        
#č na tuto třídu kladu požadavek aby kdyby něco uměla přepinat mezi zdroje
#č teď to nepotřebuji a nevím jestli to vůbec budu chtit v budoucnu.
#č Proto ji nechám teď vazanou jen na skřiňku.
class BoxEstimationData(QtCore.QObject):
    estimations_updated = QtCore.pyqtSignal(pd.DataFrame)
    
    def __init__(self, box_estimations, signal, parent=None):
        #č QObject má jediný parametr a tím je parent
        super().__init__(parent=parent)
        self.box_estimations = box_estimations
        signal.connect(self.recalculate)
        
        self.setup_context_menu()
        self.recalculate()
        
        
    def setup_context_menu(self):
        self.TRI_menu = QtWidgets.QMenu("WellMet")
        
#        self.TRI_overall_chk = QtGui.QAction("TRI_overall_estimations", self.TRI_menu) 
#        self.TRI_overall_chk.setCheckable(True)
#        self.TRI_overall_chk.setChecked(True)
#        self.TRI_overall_chk.triggered.connect(self.recalculate)
#        self.TRI_menu.addAction(self.TRI_overall_chk)
#        
#        self.simplex_chk = QtGui.QAction("Simplex estimations", self.TRI_menu) 
#        self.simplex_chk.setCheckable(True)
#        self.simplex_chk.setChecked(True)
#        self.simplex_chk.triggered.connect(self.recalculate)
#        self.TRI_menu.addAction(self.simplex_chk)
        
#        # year, it was
#        ## hope, it is temporary
#        #self.sources_action_group = QtGui.QActionGroup(self.TRI_menu)
#        #self.sources_action_group.addAction(self.TRI_overall_chk)
#        #self.sources_action_group.addAction(self.simplex_chk)
        
        #self.TRI_menu.addSeparator()
        
#        self.proxy_chk = QtGui.QAction("Proxy", self.TRI_menu) 
#        self.proxy_chk.setCheckable(True)
#        self.proxy_chk.setChecked(hasattr(self.dice_box, 'proxy'))
#        self.proxy_chk.triggered.connect(self.recalculate)
#        self.TRI_menu.addAction(self.proxy_chk)
#        
#        self.TRI_menu.addSeparator()
        
        self.reaction = QtGui.QAction("Update", self.TRI_menu)
        self.reaction.triggered.connect(self.recalculate)
        self.TRI_menu.addAction(self.reaction)
        
        self.excelaction = QtGui.QAction("Export to Excel", self.TRI_menu)
        self.excelaction.triggered.connect(self.export_to_excel)
        self.TRI_menu.addAction(self.excelaction)
        
        
    def export_to_excel(self):
        #č já bych nechtěl, aby mně export najednou spadl 
        #č z jakéhokoliv důvodu
        try:
            proposal_filename = '.xlsx'
            filename, *__ = pg.FileDialog.getSaveFileName(None, 'Export to Excel',\
                                         proposal_filename, initialFilter='*.xlsx')
            if filename:
                self.df.to_excel(filename)
        except BaseException as e:
            print(self.__class__.__name__ + ":", repr(e))
        
        
    def recalculate(self):
        self.df = df = pd.DataFrame(self.box_estimations)
        try:
            df.index = df.nsim
            #č odhady skříňky netřeba sortirovat.
            #č až se tu objeví něco dalšího, dodáme sort.
            self.estimations_updated.emit(df)
            
        except BaseException as e:
            print(self.__class__.__name__ + ":", repr(e))




class ErrorGraph(pg.PlotWidget):
    
    update_xrange = update_xrange
    
    def __init__(self, pf_exact, box_data, parent=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        
        self.pf_exact = pf_exact
        
        self.box_data = box_data
        box_data.estimations_updated.connect(self.redraw)
        
        self.setup_context_menu(box_data.TRI_menu)
        self.setup()
        self.redraw(box_data.df)
        
    def setup_context_menu(self, menu):
        # creates instance of LegendItem 
        # and saves it into plotItem.legend
        self.legend = self.addLegend()
    
        self.plotItem.ctrl.xGridCheck.setChecked(True)
        self.plotItem.ctrl.yGridCheck.setChecked(True)
        
        # menu of SimplexEstimationData
        self.plotItem.vb.menu.addMenu(menu)
        
        #č já se bojím. radší to uložím
        self.custom_menu = self.plotItem.vb.menu.addMenu("Error graph")
        
        self.legend_chk = QtGui.QAction("Legend", self.custom_menu) 
        self.legend_chk.setCheckable(True)
        self.legend_chk.triggered.connect(lambda: self.legend.setVisible(self.legend_chk.isChecked()))
        self.custom_menu.addAction(self.legend_chk)
        # apply custom menu option
        self.legend.setVisible(self.legend_chk.isChecked())

        self.laction = QtGui.QAction("Show labels", self.custom_menu)
        self.laction.triggered.connect(self.show_labels)
        self.custom_menu.addAction(self.laction)
        
            
    def show_labels(self):
        self.setLabel('left', "Failure probability estimation error")
        self.setLabel('bottom', "Number of simulations")
        
    
    def setup(self, *args, **kwargs):
        self.clear()
        self.setBackground('w')
        x = y = () # zde jen vytvoříme kostru, nakrmime daty v .redraw()
        
        # We will use logMode by default
        self.setLogMode(y=True)
        
        pen = pg.mkPen(color='m', width=2)
        self.pen_over = self.plot(x, y, pen=pen, name="pf overestimation")
        pen = pg.mkPen(color='r', width=2) #(118, 187, 255)
        self.pen_pf = self.plot(x, y, pen=pen, name="pf estimation")
        pen = pg.mkPen(color='darkMagenta', width=2) #(118, 187, 255)
        self.pen_under = self.plot(x, y, pen=pen, name="pf underestimation")
    
    
    def redraw(self, df):
        #č nevadí, ale není to slušné
#        #č zapíšeme do data rámu, snad nikomu nebude vadit
#        df['vertex_estimation_error'] = df['vertex_estimation'] - pf_exact
#        df['weighted_vertex_estimation_error'] = df['weighted_vertex_estimation'] - pf_exact
        
        if len(df) < 2: #č nevím proč hazí chyby. Asi kvůli zadané širce.
            return 
        
        if 'vertex_estimation' in df:
            v = (df['vertex_estimation'] - self.pf_exact).abs()
            self.pen_pf.setData(v.index.to_numpy(), v.to_numpy())
        if 'weighted_vertex_estimation' in df:
            wv = (df['weighted_vertex_estimation'] - self.pf_exact).abs()
            self.pen_under.setData(wv.index.to_numpy(), wv.to_numpy())
            
        xmax = df.index.max() 
        self.update_xrange(xmax*1.02)





class EstimationGraph(pg.PlotWidget):
    
    update_xrange = update_xrange
    
    def __init__(self, pf_exact, box_data, parent=None, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        
        self.pf_exact = pf_exact
        
        self.box_data = box_data
        box_data.estimations_updated.connect(self.redraw)
        
        self.setup_context_menu(box_data.TRI_menu)
        self.setup()
        self.replot()
        
    def setup_context_menu(self, menu):
        # creates instance of LegendItem 
        # and saves it into plotItem.legend
        self.legend = self.addLegend()
    
        self.plotItem.ctrl.xGridCheck.setChecked(True)
        self.plotItem.ctrl.yGridCheck.setChecked(True)
        
        # delete build-in Transforms (with Log_x and Log_y) options, 
        # they can cause uncachable exception (on any zero in data) and crash
        self.plotItem.ctrlMenu.removeAction(self.plotItem.ctrlMenu.actions()[0])
        
        #č já se bojím. radší to uložím
        self.custom_menu = self.plotItem.vb.menu.addMenu("TRI options")
        
        self.plotItem.vb.menu.addMenu(menu)
        
        self.legend_chk = QtGui.QAction("Legend", self.custom_menu) 
        self.legend_chk.setCheckable(True)
        self.legend_chk.triggered.connect(lambda: self.legend.setVisible(self.legend_chk.isChecked()))
        self.custom_menu.addAction(self.legend_chk)
        # apply custom menu option
        self.legend.setVisible(self.legend_chk.isChecked())
        
        self.log_x_chk = QtGui.QAction("Log X", self.custom_menu)
        self.log_x_chk.setCheckable(True)
        self.log_x_chk.triggered.connect(lambda: self.setLogMode(x=self.log_x_chk.isChecked()))
        self.custom_menu.addAction(self.log_x_chk)
        
        self.log_y_chk = QtGui.QAction("Log Y", self.custom_menu)
        self.log_y_chk.setCheckable(True)
        self.log_y_chk.setChecked(True)
        self.log_y_chk.triggered.connect(self.replot)
        self.custom_menu.addAction(self.log_y_chk)

        self.laction = QtGui.QAction("Show labels", self.custom_menu)
        self.laction.triggered.connect(self.show_labels)
        self.custom_menu.addAction(self.laction)
        
            
    def show_labels(self):
        self.setLabel('left', "Probability measure")
        self.setLabel('bottom', "Number of simulations")
        
    
    
#       self.legend.addItem(self.pen_success, "success domain estimation")
#        self.legend.addItem(self.pen_outside, "out of sampling domain estimation")
#        self.legend.addItem(self.pen_mix, "mixed simplices measure")
#        self.legend.addItem(self.pen_f, "failure domain estimation")
    def setup(self, *args, **kwargs):
        self.clear()
        self.setBackground('w')
        
        x = y = () # zde jen vytvoříme kostru, nakrmime daty v .redraw()
        
        #xkcd_green = (167, 255, 181) # xkcd:light seafoam green #a7ffb5
        green = (0, 255, 38, 96) 
        #xkcd_red   = (253, 193, 197) # xkcd: pale rose (#fdc1c5)
        red   = (253, 0, 17, 96)
        #xkcd_cream = (255, 243, 154) # let's try xkcd: dark cream (#fff39a)
        cream = (255, 221, 0, 96)
        grey = (196, 196, 196, 96)
        
        self.pen_f = self.plot(x, y, brush=red)#, name="failure domain estimation")
        self.pen_f.setZValue(-100)
        
        self.pen_success = self.plot(x, y, brush=green) #, name="success domain estimation") 
        self.pen_success.setZValue(-100)
        
        self.pen_outmix = self.plot(x, y)
        
        self.fill_mix = pg.FillBetweenItem(self.pen_f, self.pen_outmix)
        #self.fill_mix.setData(name="mixed simplices measure")
        self.fill_mix.setBrush(cream)
        self.fill_mix.setZValue(-100)
        self.addItem(self.fill_mix)
        
        #self.pen_outside = self.plot(x, y)
        self.fill_outside = pg.FillBetweenItem(self.pen_outmix, self.pen_success)
        #self.fill_outside.setData(name="out of sampling domain estimation")
        self.fill_outside.setBrush(grey)
        self.fill_outside.setZValue(-100)
        self.addItem(self.fill_outside)
        
        self.one_ruler = self.addLine(y=1, pen='k') 
        self.zero_ruler = self.addLine(y=0, pen='k') 
            
        if self.pf_exact > 0:
            exact_name = "exact solution"
            pen = pg.mkPen(color='b', width=1.5) # blue
            self.pen_exact = self.addLine(y=self.pf_exact, pen=pen, name=exact_name) 
            #č aby se nám něco zobrazovalo v legendu
            self.pen_exact_PR = self.plot(x, y, pen=pen, name=exact_name)
            
        pen = pg.mkPen(color='grey', width=2)
        self.pen_outside = self.plot(x, y, pen=pen, name="outside probability")
        pen = pg.mkPen(color='orange', width=2)
        self.pen_mixed = self.plot(x, y, pen=pen, name="mixed probability")
        
        pen = pg.mkPen(color='m', width=2)
        self.pen_over = self.plot(x, y, pen=pen, name="pf overestimation")
        pen = pg.mkPen(color='r', width=2) #(118, 187, 255)
        self.pen_pf = self.plot(x, y, pen=pen, name="pf estimation")
        pen = pg.mkPen(color='darkMagenta', width=2) #(118, 187, 255)
        self.pen_under = self.plot(x, y, pen=pen, name="pf underestimation")
        
        
    def replot(self, *args, **kwargs):
        if self.log_y_chk.isChecked():
            self.one_ruler.hide()
            
            if self.pf_exact > 0:
                self.pen_exact.setPos(np.log10(self.pf_exact))
                
                
            self.setLogMode(y=True)
            #self.pen_f.setPen(pg.mkPen(color=(255, 0, 0), width=3)) #, style=QtCore.Qt.DashLine)
            self.pen_f.setPen(None)
            self.pen_f.setFillLevel(None)
            self.pen_success.setFillLevel(0)
            
        else:
            self.one_ruler.show()
            
            if self.pf_exact > 0:
                self.pen_exact.setPos(self.pf_exact)
            
                
            self.setLogMode(y=False)
            self.pen_f.setPen(None)
            self.pen_f.setFillLevel(0)
            self.pen_success.setFillLevel(1)
        
        self.redraw(self.box_data.df)
        
    @staticmethod
    def _set_data(plot_item, serie):
        plot_item.setData(serie.index.to_numpy(), serie.to_numpy())


    def redraw(self, df):
        if len(df) < 2: #č nevím proč hazí chyby. Asi kvůli zadané širce.
            return 
        
        if 'vertex_ratio_estimation' in df:
            self._set_data(self.pen_over, df['vertex_ratio_estimation'])
        
        if 'vertex_estimation' in df:
            self._set_data(self.pen_pf, df['vertex_estimation'])
        
        if 'weighted_vertex_estimation' in df:
            self._set_data(self.pen_under, df['weighted_vertex_estimation'])
        if 'weighted_ratio_estimation' in df:
            self._set_data(self.pen_under, df['weighted_ratio_estimation'])
        
        self._set_data(self.pen_outside, df.outside)
        self._set_data(self.pen_mixed, df.mix)
        
        #č outside by neměl klesnout do nuly.
        #č aj kdyby něco - jen se pokazí maliňko zalivky
        success = df.failure + df.mix + df.outside
        outmix = df.failure + df.mix
        failure = df.failure
        
        #č qtpygraph už nespadne kvuli nule se zapnutým LogModem
        #č ale pořad se to hodí pro zalivky.
        #č Jsou vazany na pen_f, pen_outmix a pen_success
        # (in case of LogPlot) fallback values also used
        if self.log_y_chk.isChecked():
            outmix.where(outmix > 0, success, inplace=True)
            failure = failure.where(failure > 0, outmix)
        
        self._set_data(self.pen_success, success)
        self._set_data(self.pen_f, failure)
        self._set_data(self.pen_outmix, outmix)
        
        xmax = df.index.max() 
        self.update_xrange(xmax*1.02)

            


#
# DEPRECATED
# 


def get_estimation_data(estimations, metric):
    metric_dict = dict()
    # new-style: šecko leží dohromady a každý z toho
    # bere co chce a jak chce
    # ne že by to bylo nějak šetrný
    # estimation je slovníkem
    for estimation in estimations:
        # nsim musí mäť každej odhad
        # pokud nemá - je třeba jej prostě opravit
        nsim = estimation['nsim']
        try: 
            metric_dict[nsim] = estimation[metric]
        except KeyError as e:
            pass #print(self.__class__.__name__ + ":", repr(e))
    
    # nikdo neslibil, že budou v pořadí
    x = np.sort(tuple(metric_dict.keys()))
    y = np.array(tuple(metric_dict.values()))[np.argsort(tuple(metric_dict.keys()))]
    return x, y


class SimplexEstimationData(QtCore.QObject):
    #š budeme mӓť svůj vlastní signaľčík
    estimations_updated = QtCore.pyqtSignal(pd.DataFrame)
    
    def __init__(self, dice_box, stream=None, *args, **kwargs):
        super().__init__(stream, *args, **kwargs)
        self.dice_box = dice_box
        #č je zřejmě, že tím potokem bylo myšleno hlavní okínko
        #č asi aby nepadalo, když nenajde signaly
        self.stream = stream
        if stream is not None:
            self.stream.box_runned.connect(self.recalculate)
            self.stream.estimation_added.connect(self.recalculate)
        
        self.setup_context_menu()
        self.recalculate()
        
        
    def setup_context_menu(self):
        # simplex_data_menu
        self.TRI_menu = QtWidgets.QMenu("TRI sources", self.stream)
        
        self.TRI_overall_chk = QtGui.QAction("TRI_overall_estimations", self.TRI_menu) 
        self.TRI_overall_chk.setCheckable(True)
        self.TRI_overall_chk.setChecked(True)
        self.TRI_overall_chk.triggered.connect(self.recalculate)
        self.TRI_menu.addAction(self.TRI_overall_chk)
        
        self.simplex_chk = QtGui.QAction("Simplex estimations", self.TRI_menu) 
        self.simplex_chk.setCheckable(True)
        self.simplex_chk.setChecked(True)
        self.simplex_chk.triggered.connect(self.recalculate)
        self.TRI_menu.addAction(self.simplex_chk)
        
        # year, it was
        ## hope, it is temporary
        #self.sources_action_group = QtGui.QActionGroup(self.TRI_menu)
        #self.sources_action_group.addAction(self.TRI_overall_chk)
        #self.sources_action_group.addAction(self.simplex_chk)
        
        self.TRI_menu.addSeparator()
        
        self.proxy_chk = QtGui.QAction("Proxy", self.TRI_menu) 
        self.proxy_chk.setCheckable(True)
        self.proxy_chk.setChecked(hasattr(self.dice_box, 'proxy'))
        self.proxy_chk.triggered.connect(self.recalculate)
        self.TRI_menu.addAction(self.proxy_chk)
        
        self.TRI_menu.addSeparator()
        
        self.reaction = QtGui.QAction("Update", self.TRI_menu)
        self.reaction.triggered.connect(self.recalculate)
        self.TRI_menu.addAction(self.reaction)
        
        self.excelaction = QtGui.QAction("Export to Excel", self.TRI_menu)
        self.excelaction.triggered.connect(self.export_to_excel)
        self.TRI_menu.addAction(self.excelaction)
        
        
    def export_to_excel(self):
        #č já bych nechtěl, aby mně export najednou spadl 
        #č z jakéhokoliv důvodu
        try:
            proposal_filename = self.dice_box.guessbox.filename
            if proposal_filename:
                proposal_filename += '.xlsx'
            else:
                proposal_filename = self.dice_box.gm_signature + '.xlsx'
            filename, *__ = pg.FileDialog.getSaveFileName(self.stream, 'Export to Excel',\
                                         proposal_filename, initialFilter='*.xlsx')
            self.df.to_excel(filename)
        except BaseException as e:
            print(self.__class__.__name__ + ":", repr(e))
        
        
    def recalculate(self):
        try:
            # sources=['box', 'user']
            sources = list()
            if self.TRI_overall_chk.isChecked():
                sources.append('box')
            if self.simplex_chk.isChecked():
                sources.append('user')
                
            self.df = stm_df.get_tri_data_frame(self.dice_box, sources=sources,\
                                        apply_proxy=self.proxy_chk.isChecked())
            self.estimations_updated.emit(self.df)
            
        except BaseException as e:
            print(self.__class__.__name__ + ":", repr(e))
        



        






class VoronoiEstimationGraph(pg.PlotWidget):
    def __init__(self, black_box, samplebox_item, parent=None, *args, **kwargs):
        super().__init__(parent)
        self.sb_item = samplebox_item
        self.sb_item.box_runned.connect(self.redraw)
        self.sb_item.estimation_added.connect(self.redraw)
        
        self.black_box = black_box
        self.setBackground('w')
        self.setLimits(xMin=-0.45)
        
        self.reaction = QtGui.QAction("Redraw", self.plotItem.ctrlMenu)
        self.reaction.triggered.connect(self.redraw)
        self.plotItem.ctrlMenu.insertAction(self.plotItem.ctrlMenu.actions()[0], self.reaction)
        
        
        
        # implicitně Y je v logaritmickem měřítku
        self.setLogMode(False, True)

        x = y = () # zde jen vytvoříme kostru, nakrmíme daty v .redraw()
        
        
        # nechapu, proč těm Itemům ríkám "propíska" 
        # propíska? Их есть у нас!
        
        self.Voronoi_2_point_upper_bound = self.plot(x, y, pen='y')
        self.Voronoi_2_point_lower_bound = self.plot(x, y, pen='y')
        
        fill_color = (255, 243, 154) # let's try xkcd: dark cream (#fff39a)
        self.fill = pg.FillBetweenItem(self.Voronoi_2_point_upper_bound, self.Voronoi_2_point_lower_bound, fill_color)
        self.addItem(self.fill)
        
        self.Voronoi_2_point_failure_rate = self.plot(x, y, pen=(195,46,212))
        self.Voronoi_2_point_pure_failure_rate = self.plot(x, y, pen='m')
        self.Voronoi_failure_rate = self.plot(x, y, pen='r')
        
        self.pen_exact = self.plot(x, y, pen='b') # blue
        self.pen_one = self.plot(x, y, pen='k') # black
        
        self.redraw()
    
   
    def redraw(self):
        # kruci, ještě navic i generovať pokažde znovu...
        metrics = {'Voronoi_2_point_upper_bound':{},\
                    'Voronoi_2_point_lower_bound':{},\
                    'Voronoi_2_point_failure_rate':{},\
                    'Voronoi_2_point_pure_failure_rate':{},\
                    'Voronoi_failure_rate':{},}
        xmin = np.inf
        xmax = -np.inf
        try: # тут всё что угодно может пойти не так
            # new-style: šecko leží dohromady a každý z toho
            # bere co chce a jak chce
            # ne že by to bylo nějak šetrný
            # estimation je slovníkem
            for estimation in self.black_box.estimations:
                # nsim musí mäť každej odhad
                # pokud nemá - je třeba jej prostě opravit
                nsim = estimation['nsim']
                
                
                for metric, metric_dict in metrics.items():
                    try: 
                        if estimation[metric] > 0:
                            metric_dict[nsim] = estimation[metric]
                            if nsim > xmax:
                                xmax = nsim
                            if nsim < xmin:
                                xmin = nsim
                    except KeyError as e:
                        pass #print(self.__class__.__name__ + ":", repr(e))
            
            for metric, metric_dict in metrics.items():
                pen = getattr(self, metric)
                # nikdo neslibil, že budou v pořadí
                x = np.sort(tuple(metric_dict.keys()))
                y = np.array(tuple(metric_dict.values()))[np.argsort(tuple(metric_dict.keys()))]
                pen.setData(x, y)
                
            if (xmax - xmin) > 0:
                self.pen_one.setData((xmin,xmax), (1, 1))
                if hasattr(self.black_box, 'pf_exact'):
                    # poslední. I když spadne, tak už nikomu moc nevadí
                    self.pen_exact.setData((xmin,xmax), (self.black_box.pf_exact, self.black_box.pf_exact))
        
        except BaseException as e:
            print(self.__class__.__name__ + ":", repr(e))
        
        
        # pen_f.opts['logMode']
        # pen_outside.setLogMode(False, False)
        #setLogMode(False, False)
        #f = pg.FillBetweenItem(curves[i], curves[i+1], brushes[i])
        #win.addItem(f)










#!/usr/bin/env python
# coding: utf-8


#č nazvy proměnných jsou v angličtině
#č Ale komenty teda ne)

#č otázkou je, kdo je uživatelem modulu.
#č zatím ho potřebujou jenom vnejší skripty, 
#č které dostávájí data z csv souborů a žádné skřiňky nevytvařejí.
#č Proto funkce v tomto modulu zatím-prozátím nebudou spolehát na
#č ax.sample_box

from scipy import stats # for tri_beta_plot

#č sehnaní datarámu necháme uživateli
def tri_estimation_fill(ax, df):
    #xkcd_green = (167, 255, 181) # xkcd:light seafoam green #a7ffb5
    green = "#A7FFB5"
    #xkcd_red   = (253, 193, 197) # xkcd: pale rose (#fdc1c5)
    red   = "#FDC1C5"
    #xkcd_cream = (255, 243, 154) # let's try xkcd: dark cream (#fff39a)
    cream = "#FFF39A"
    grey = "#DDDDDD"
    
    o = df['outside'].to_numpy()
    s = df['success'].to_numpy()
    f = df['failure'].to_numpy()
    m = df['mix'].to_numpy()
    kwargs = {'colors': (red, cream, grey, green),
            'labels': ("failure domain", "mixed domain",\
             "outside domain", "success domain")}
    if len(df.index) > 1000:
        kwargs['rasterized'] = True
    return ax.stackplot(df.index, f, m, o, s, **kwargs)


#č sehnaní datarámu necháme uživateli
#
#č musím trochu davat bacha,
#č externí skripta df-ko vytvařejí sámi jak chcou
#č proto se nemá obecně spolehat na korektní index
def shell_estimation_fill(ax, df, use_df_index=False):
    if use_df_index:
        x = df.index
    else:
        x = df['nsim'].to_numpy()
    
    shell_color = "#ECCEAF"  #"#FFFD95"
    outer_color = "#CCCCCC"
    colors = (outer_color, shell_color)
    outer = df['outer'].to_numpy()
    shell = outer + df['shell'].to_numpy()
    labels = ("outer", "annulus")
    return ax.stackplot(x, outer, shell, labels=labels, colors=colors)


#č sehnaní datarámu necháme uživateli
# inverted
def trii_estimation_fill(ax, df):
    #xkcd_green = (167, 255, 181) # xkcd:light seafoam green #a7ffb5
    green = "#A7FFB5"
    #xkcd_red   = (253, 193, 197) # xkcd: pale rose (#fdc1c5)
    red   = "#FDC1C5"
    #xkcd_cream = (255, 243, 154) # let's try xkcd: dark cream (#fff39a)
    cream = "#FFF39A"
    grey = "#DDDDDD"
    colors = (grey, red, cream, green)
    o = df['outside'].to_numpy()
    s = df['success'].to_numpy()
    f = df['failure'].to_numpy()
    m = df['mix'].to_numpy()
    labels = ("out of sampling domain estimation", "failure domain estimation",\
             "mixed simplices measure", "success domain estimation")
    return ax.stackplot(df.index, o, f, m, s, labels=labels, colors=colors)


#č ok, tak uděláme všecko dohromady
#č datarám, jako vždy, uživatel donese svůj vlastní
def tri_estimation_plot(ax, df, pf_exact=None, pf_exact_method="$p_f$",
                         plot_outside=True, plot_mix=True, **kwargs):
    # some default values
#    if not ax.get_xlabel():
#        ax.set_xlabel('Number of simulations')
#    if not ax.get_ylabel():
#        ax.set_ylabel('Probability measure')
        
    # fill
    tri_estimation_fill(ax, df)
    
    #č blbost, ale uspořadal jsem tu prvky tak,
    #č aby se hezky kreslily v legendě
    
    if (len(df.index) > 1000) and ('rasterize' not in kwargs):
        kwargs['rasterized'] = True
    
    lw = kwargs.pop('lw', kwargs.pop('linewidth', 1.5))
    
    try:
        v = df['vertex_estimation'].to_numpy()
        #vr = df['vertex_ratio_estimation'].to_numpy()
        #wr = df['weighted_ratio_estimation'].to_numpy()
        #ax.plot(df.index, vr, '-m', label="$p_f$ vertex ratio estimation", zorder=10500, lw=lw/2, **kwargs)
        #ax.plot(df.index, wr, '-', label="$p_f$ wegthed ratio estimation", color='darkmagenta', zorder=10500, lw=lw/2, **kwargs)
        ax.plot(df.index, v, '-r', label="simple $p_f$ estimation", zorder=100500, lw=lw, **kwargs)
    except:
        pass
#        v = df['vertex_estimation'].to_numpy()
#        wv = df['weighted_vertex_estimation'].to_numpy()
#        ax.plot(df.index, wv, '-r', label="weighted $p_f$ estimation", zorder=100500, **kwargs)
#        ax.plot(df.index, v, '-m', label="simple $p_f$ estimation", zorder=10500, **kwargs)
    
    
    #č teď čáry
    if plot_outside:
        o = df['outside'].to_numpy()
        ax.plot(df.index, o, '-', color="#AAAAAA", \
                label="outside domain estimation", zorder=150, **kwargs)
                
    if plot_mix:
        m = df['mix'].to_numpy()
        ax.plot(df.index, m, '-', color="#FF8000", \
                label="mixed domain estimation", zorder=1050, **kwargs)
    
    if pf_exact is not None:
        ax.axhline(pf_exact, c='b', label=pf_exact_method, **kwargs)
    

#č ok, tak uděláme všecko dohromady
#č datarám, jako vždy, uživatel donese svůj vlastní
def tri_beta_plot(ax, df, pf_exact=None, pf_exact_method="$p_f$",
                         plot_outside=True, plot_mix=True, **kwargs):
    
    #č blbost, ale uspořadal jsem tu prvky tak,
    #č aby se hezky kreslily v legendě
    
    v = -stats.norm.ppf(df['vertex_estimation'].to_numpy())
    #wv = -stats.norm.ppf(df['weighted_vertex_estimation'].to_numpy())
    #ax.plot(df.index, wv, '-r', label="weighted $p_f$ estimation", zorder=100500, **kwargs)
    ax.plot(df.index, v, '-r', label="vertex $p_f$ estimation", zorder=10500, **kwargs)
    
    
    #č teď čáry
    if plot_outside:
        o = -stats.norm.ppf(df['outside'].to_numpy())
        ax.plot(df.index, o, '-', color="#AAAAAA", \
                label="outside domain estimation", zorder=150, **kwargs)
                
    if plot_mix:
        m = -stats.norm.ppf(df['mix'].to_numpy())
        ax.plot(df.index, m, '-', color="#FF8000", \
                label="mixed domain estimation", zorder=1050, **kwargs)
    
    if pf_exact is not None:
        ax.axhline(-stats.norm.ppf(pf_exact), c='b', label=pf_exact_method, **kwargs)

    

#č ok, tak uděláme všecko dohromady
#č datarám, jako vždy, uživatel donese svůj vlastní
def shell_estimation_plot(ax, df, use_df_index=False, **kwargs):
    
    # fill
    shell_estimation_fill(ax, df, use_df_index)
    
    #č teď čáry
    if use_df_index:
        x = df.index
    else:
        x = df['nsim'].to_numpy()
    
    #č je to ten opravdový outside z tri plot 
    #č ponecháme i jeho původní barvu
    o = df['outside'].to_numpy()
    ax.plot(x, o, '-', color="#AAAAAA", \
            label="outside estimation", **kwargs)
    
    y = df['FORM_outside'].to_numpy()
    ax.plot(x, y, '-', color="tab:purple", \
            label="hyperplane outside approximation", **kwargs)
    
    y = df['2FORM_outside'].to_numpy()
    ax.plot(x, y, '-', color="tab:pink", \
            label="two hyperplanes outside approximation", **kwargs)
    
    y = df['orth_outside'].to_numpy()
    ax.plot(x, y, '-', color="tab:brown", \
            label="hypercube outside approximation", **kwargs)


#č ok, tak uděláme všecko dohromady
#č datarám, jako vždy, uživatel donese svůj vlastní
def trii_estimation_plot(ax, df, pf_exact=None, pf_exact_method="$p_f$"):
    # some default values
#    if not ax.get_xlabel():
#        ax.set_xlabel('Number of simulations')
#    if not ax.get_ylabel():
#        ax.set_ylabel('Probability measure')
    # fill
    trii_estimation_fill(ax, df)
    
    #č teď čáry
    v = df['vertex_estimation'].to_numpy()
    wv = df['weighted_vertex_estimation'].to_numpy()
    # v trii grafu máme outside zdolu
    mask = v > 0
    o = df['outside'].to_numpy()[mask]
    vo = v[mask] + o
    wvo = wv[mask] + o
    ax.plot(df.index[mask], vo, '-m', label="simple $p_f$ estimation")
    ax.plot(df.index[mask], wvo, '-r', label="weighted $p_f$ estimation")
    
    if pf_exact is not None:
        ax.axhline(pf_exact, c='b', label=pf_exact_method)
    
    
        
##xkcd_green = (167, 255, 181) # xkcd:light seafoam green #a7ffb5
#green = (0, 255, 38, 96) 
##xkcd_red   = (253, 193, 197) # xkcd: pale rose (#fdc1c5)
#red   = (253, 0, 17, 96)
##xkcd_cream = (255, 243, 154) # let's try xkcd: dark cream (#fff39a)
#cream = (255, 221, 0, 96)
#grey = (196, 196, 196, 96)
        
class SimplexErrorGraph:
        
            
    def show_labels(self):
        self.setLabel('left', "Failure probability estimation error")
        self.setLabel('bottom', "Number of simulations")
        
    
    def setup(self, *args, **kwargs):
        
        #xkcd_red   = (253, 193, 197) # xkcd: pale rose (#fdc1c5)
        #red   = (253, 0, 17, 96)
        
        #self.pen_f = self.plot(x, y, brush=red)#, name="failure domain estimation")
        #self.pen_f.setZValue(-100)
        
        
        pen = pg.mkPen(color='m', width=2)
        self.pen_vertex = self.plot(x, y, pen=pen, name="simple pf estimation")
        pen = pg.mkPen(color='r', width=2) #(118, 187, 255)
        self.pen_weighted_vertex = self.plot(x, y, pen=pen, name="weighted pf estimation")
        
        
    
    #č když se někde objeví nula se zapnutým LogModem - 
    #č qtpygraph hned spadne a není možne ten pad zachytit
    def zerosafe(self, x, y, fallback_y=None): 
        x = np.array(x)
        y = np.array(y)
        if fallback_y is None:
            fallback_y = y
        y = np.where(y > 0, y, fallback_y)
        mask = y > 0
        return x[mask], y[mask]
        
    
    def redraw(self):
        #č neotravujme uživatele chybovejma hlaškama
        if hasattr(self.simplex_data.dice_box, 'pf_exact'):
            try: #ё тут всё что угодно может пойти не так
                pf_exact = self.simplex_data.dice_box.pf_exact
                
                df = self.simplex_data.df
                #č zapíšeme do data rámu, snad nikomu nebude vadit
                df['vertex_estimation_error'] = df['vertex_estimation'] - pf_exact
                df['weighted_vertex_estimation_error'] = df['weighted_vertex_estimation'] - pf_exact
                
                v = df['vertex_estimation_error'].abs()
                wv = df['weighted_vertex_estimation_error'].abs()
                
                x, y = self.zerosafe(v.index, v.to_numpy())
                self.pen_vertex.setData(x, y)
                
                x, y = self.zerosafe(wv.index, wv.to_numpy())
                self.pen_weighted_vertex.setData(x, y)
            
                
            except BaseException as e:
                print(self.__class__.__name__ + ":", repr(e))





#!/usr/bin/env python
# coding: utf-8

#оӵ Тӧдьы кабъёс (ящикьёс) эскером

from .. import whitebox
from .. import samplebox #č pro proxy_prod
from .. import g_models as gm
from .. import f_models
import scipy.stats as stats

import numpy as np

f = f_models.SNorm(2)
h = f_models.UnCorD((stats.norm(-12,5), stats.norm(12,5)))
#l = f_models.UnCorD([stats.lognorm(s=1), stats.lognorm(s=1)])
#č pro různorodost
gew = f_models.Nataf((stats.gumbel_r, stats.weibull_min(c=1.5)), [[1,0.8], [0.8,1]])
gewm = f_models.Nataf((stats.gumbel_r, stats.weibull_min(c=1.5)), [[1,-0.8], [-0.8,1]])
gb = f_models.Nataf((stats.gumbel_r(scale=0.3), stats.beta(0.6, 0.8)), [[1,-0.75], [-0.75,1]])
#gew = f_models.Nataf((stats.norm, stats.norm), [[1,0.8], [0.8,1]])
#gew = f_models.Nataf((stats.gumbel_r, stats.gumbel_l), [[1,0.8], [0.8,1]])
#č uniforma je na [0, 1], takže v cajku
u = f_models.UnCorD((stats.uniform, stats.uniform))

__all__ = []

#č zde ten __all__ těžko dohlídáme
#č budeme přidávat do seznamu postupně
def add(str):
    __all__.append(str)


# HyperRovina, куда ж без неё?
#############
# g= a*X1 + b*X2 + c
# becames
# gm = Linear_nD(betas=(a,b,c))
add('natafm_plane_2D')
def natafm_plane_2D():
    wt = whitebox.WhiteBox(gewm, gm.Linear_nD((-1, -2, 7)))
    wt.pf_exact = 0.00114
    wt.Nsim = int(10e8)
    wt.pf_exact_method = 'MC'
    return wt



# nobody knows whats going on there
""" parametry pravdepodobnostniho rozdeleni pro Z_min s Weib. velicinami
wb_scales=(1,1) - tuple of Weibull scale parameters, len(wb_scales)==nvar
shape = 5 
je třeba zadat buď pf_exact, nebo konštantu u funkce minima Z_min """
# g= min(X1, X2, X3, X4) + c
add('weibull_min_2D')
def weibull_min_2D():
    return whitebox.Weibull_Z_min(wb_scales=(1.5,0.9), shape=4, const=-2.5)
    #whitebox.Weibull_Z_min(wb_scales=(1.5,0.9), shape=4, pf_exact=1.0e-6)



# neither Gauß knows whats going on there
#whitebox.Gaussian_Z_sumexp(nvar=2, pf_exact=1.0e-6)




# je třeba zadat buď pf_exact, nebo konštantu u funkce Z_sumsq
# g = np.sum(sample**2, axis=1) - self._const
add('sumsq_2D')
def sumsq_2D():
    return whitebox.SNorm_Z_sumsq(nvar=2, const=9.2)
    #whitebox.SNorm_Z_sumsq(nvar=2, pf_exact=1.0e-6)



# libustka
add('ball_2D')
def ball_2D():
    return whitebox.SNorm_S_ball(nvar=2, r=5)




"""suma velicin plus beta*sqrt(Nvar. )
Pro IID Gaussian ma tohle ind. spol. beta = beta
The same as Linear_nD, but defined via 
beta in sense of reliability index"""
# Z_sum s divokým rozdělením
# pro obyčejný norm dist viz. HyperPlane
add('nataf_sum_2D')
def nataf_sum_2D():
    wt = whitebox.WhiteBox(gew, gm.Z_sum(nvar=2, beta_exact=np.sqrt(2)))
    wt.pf_exact = 0.000495
    wt.Nsim = int(10e6)
    wt.pf_exact_method = 'MC'
    return wt


# soucin velicin plus nějaká konstanta 
# g= X1 * X2 * X3 * X4 + c
add('prod_2D_1')
add('prod_2D_5')
def prod_2D_1():
    return whitebox.WhiteBox(f, gm.Z_prod(const=1))
def prod_2D_5():
    return whitebox.WhiteBox(f, gm.Z_prod(const=5))

gm_z_prod = gm.Z_prod(const=5)

def proxy_prod(input_sample):
    # očekávam, že get_R_coordinates mně vrátí 2D pole
    sample = gm.get_R_coordinates(input_sample, 2)
    x, y = sample.T
    
    # osudná podmínka
    mask = np.atleast_1d(np.sign(x)==np.sign(y)).astype(bool)
    #č mrdáme na kontrolu. současný startup candybox vytvoří vždycky
    input_sample.candybox.proxy = mask
    sweet_sample = input_sample
    #sweet_sample = CandyBox(input_sample, proxy=mask)
            
    # zatím, pro jednoduchost, předpokladáme,
    # že dostaváme vzorky po jednom
    if np.all(mask):
        # nevíme vůbec, co to je za funkci
        # ale veříme, že víme co tam bude
        g = np.full(len(sweet_sample), 1)
        # s praznejm podpísem odmítá
        return samplebox.SampleBox(sweet_sample, g, 'proxy_prod')
        
    else: # deme počítat, bez b
        true_sample = gm_z_prod(sweet_sample)
        # padelame pospís
        true_sample.gm_signature = 'proxy_prod'
        return true_sample

# wrap
proxy_prod.get_2D_R_boundary = gm_z_prod.get_2D_R_boundary

add('proxy_prod_2D_5')
def proxy_prod_2D_5():
    return whitebox.WhiteBox(f, proxy_prod)


# min velicin plus nějaká konstanta 
# g= min(X1, X2, X3, X4) + c
add('min_2D')
def min_2D():
    return whitebox.Gaussian_Z_min(ndim=2, const=5)



# g = np.sum(np.exp(-(sample**2)), axis=1) + self._const
add('nataf_supexp_2D')
def nataf_supexp_2D():
    return whitebox.WhiteBox(gew, gm.Z_sumexp(const=-0.003))
add('nataf_sumexp_2D')
def nataf_sumexp_2D():
    wt = whitebox.WhiteBox(gewm, gm.Z_sumexp(const=-0.003))
    wt.pf_exact = 1.2e-05
    wt.Nsim = int(10e6)
    wt.pf_exact_method = 'MC'
    return wt


"""
c = 0.5 # wave amplitude in Gaussian space
d = 3.0 # average of sine fiunction in Gaussian space
k = 6   # number of sine waves (design points)
"""
add('sinball_2D')
def sinball_2D():
    return whitebox.WhiteBox(f, gm.S_ballSin2D(c=0.5, d=3.0, k=6))





# g = np.sum(sample**2, axis=1) - self._const
add('nataf_sumsq_2D')
def nataf_sumsq_2D():
    return whitebox.WhiteBox(gew, gm.Z_sumsq(const=9.2))




# obecný S_ball bez výpočtu pf
add('nataf_ball_2D')
def nataf_ball_2D():
    wt = whitebox.WhiteBox(gew, gm.S_ball(r=5))
    wt.pf_exact = 0.012403
    wt.Nsim = int(10e6)
    wt.pf_exact_method = 'MC'
    return wt

add('natafm_ball_2D')
def natafm_ball_2D():
    wt = whitebox.WhiteBox(gewm, gm.S_ball(r=5))
    wt.pf_exact = 0.006812
    wt.Nsim = int(10e6)
    wt.pf_exact_method = 'MC'
    return wt
    
add('beta_circle_2D')
def beta_circle_2D():
    wt = whitebox.WhiteBox(gb, gm.ConicSection(l=1.5, e=0, c=(0.5,0), sign=-1))
    wt.pf_exact = 0.001622
    wt.Nsim = int(10e6)
    wt.pf_exact_method = 'MC'
    return wt


# kruznicka
add('circle_2D')
def circle_2D():
    return whitebox.WhiteBox(f, gm.ConicSection(l=2, e=0, teta=-np.pi/4, c=(-3,1), sign=1))


# Fajvka
# člověk se tu bez půllitry nevýzná
# viz. conic_section_boundaries_test.py pro inspiraciju
add('five_2D')
def five_2D():
    return whitebox.WhiteBox(h, gm.ConicSection(l=.1, e=1.1, teta=-np.pi/4, c=(-3,1), sign=1))


# Exp_P
# g = y - 1./np.exp(x)**5
# g = y - self._k/np.exp(x)**self._pow
# k tomuhlenstomu by bylo možně výtvořiť bílou skřiňku (pro výpočet pf)
add('exp_P_2D')
def exp_P_2D():
    return whitebox.WhiteBox(u, gm.Exp_P(k=1., pow=5))

add('nataf_exp_2D')
def nataf_exp_2D():
    return whitebox.WhiteBox(gew, gm.Exp_P(k=1., pow=-0.5))


# Sin2D
# g = self._kx * x + self._ky * y + np.sin(self._kxsin*x) + self._const
add('sin_2D')
def sin_2D():
    return whitebox.WhiteBox(f, gm.Sin2D(kx=-1/4., ky=-1, kxsin=5, const=5))





# Prod_FourBetas
# g = beta^2/2 - |x1 * x2|
#whitebox.WhiteBox(f, gm.Prod_FourBetas(beta=2.0))



""" BlackSwan2D
a = 2.0 # boundary for x1
b = 5.0 # boundary for x2
y = np.where(sim[:,0] <= a, sim[:,0], sim[:,1])
# pro x1 <= a   y = x1
# pro x1 >  a   y = x2
g = b - y # failure for b<y"""
#whitebox.WhiteBox(f, gm.BlackSwan2D(a=2.0, b=5.0))



""" Metaballs2D
# sebemenší parametrizace
y1 = 4/9*(x1 + 2  )**2 + 1/25 * (x2    )**2 
y2 = 1/4*(x1 - 2.5)**2 + 1/25 * (x2-0.5)**2 
g = 30.0/( y1**2 + 1.0 ) + 20.0/( y2**2 + 1.0 ) - self._const"""
#whitebox.WhiteBox(f, gm.Metaballs2D(const=5))




""" Logistic2D
# sebemenší parametrizace
y1 = self._c1 - x1
y2 = self._c2 + x2
y3 = 1.0/(1+np.exp(-2.0*y2)) - 0.5

if self.easy_version:
    g = np.minimum(y1,y2)  # easy version for SuS
else:
    g = np.minimum(y1,y3)  # difficult version for SuS"""
#whitebox.WhiteBox(f, gm.Logistic2D(c1=5, c2=4, easy_version=True))




""" CosExp2D
# sebemenší parametrizace
s = self._s
# g = cos((np.exp(-xm-s  ))*xm)   * np.exp(-(x +s  )/3)
g = np.cos( ( np.exp(-sample[:,0] - s ) )*sample[:,0])   * np.exp( -(sample[:,0] + s  )/3 )  """
#whitebox.WhiteBox(f, gm.CosExp2D(s=5))



""" FourBranch2D
g1 = k1 + 0.1*(x1 - x2)**2 - (x1 + x2)/np.sqrt(2)
g2 = k1 + 0.1*(x1 - x2)**2 + (x1 + x2)/np.sqrt(2)
g3 = (x1 - x2) + k2/np.sqrt(2)
g4 = (x2 - x1) + k2/np.sqrt(2) #č byl tu překlep v jednom članku
g = np.min((g1, g2, g3, g4), axis=0)"""
#whitebox.WhiteBox(f, gm.FourBranch2D(k1=3, k2=7))






#!/usr/bin/env python
# coding: utf-8

import pyqtgraph as pg
import pyqtgraph.opengl as gl
from pyqtgraph.Qt import QtWidgets
from pyqtgraph.Qt import QtCore


import numpy as np
import pandas as pd # required for estimation graph

from scipy import spatial

from .. import estimation as stm
from .. import simplex as six
from . import qt_gui
from . import qt_plot


### Define a top-level widget to hold everything
class QtGuiPlot3D(qt_gui.QtGuiWindow):
    #sb_runned = QtCore.pyqtSignal()
    
    def initialize_central_widget(self):
        self.central_widget = gl.GLViewWidget()
        self.redraw_called.connect(self.redraw_plot)
        
        self.base_plotting = BasePlotting(self)
        self.px_size = 5

    def redraw_plot(self):
        try:
            self.central_widget.clear()
            self.central_widget.reset()
        except: #č 0.10 ještě neměla ty funkce
            for item in self.central_widget.items:
                item._setView(None)
            self.central_widget.items = []
            self.central_widget.update()
        self.central_widget.opts['fov'] = 3

        

    def plot_setup(self):
        self.view_items = []
        self.view_items.append(UnitCube(self))
        self.view_items.append(Axes(self))
        self.view_items.append(Grid(self))
        self.view_items.append(LastShot(self))
        self.view_items.append(Density(self))
        self.view_items.append(ConvexHull(self))
        self.view_items.append(Facets(self))
        self.view_items.append(Wireframe(self))
        

        dock = dock_r = QtWidgets.QDockWidget("Simplex-based pf estimation", self)
        dock.setWidget(SimplexEstimationWidget(self, dock))
        self.view.addAction(dock.toggleViewAction())
        self.dockables.append(dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)
        
        #!
        dock = QtWidgets.QDockWidget("Tesselation-based pf estimation", self)
        dock.setWidget(VoronoiEstimationWidget(self, dock))
        self.view.addAction(dock.toggleViewAction())
        self.dockables.append(dock)
        #self.addDockWidget(QtCore.Qt.RightDockWidgetArea, dock)
        self.tabifyDockWidget(dock_r, dock)
        
        
        dock = QtWidgets.QDockWidget("Blackbox's candidates", self)
        dock.setWidget(CandidatesWidget(self, dock))
        self.view.addAction(dock.toggleViewAction())
        self.dockables.append(dock)
        self.tabifyDockWidget(dock_r, dock)


        
        



class BasePlotting(gl.GLScatterPlotItem):
    def __init__(self, w):
        super().__init__()
        self.w = w
        #self.w.box_runned.connect(self.redraw) #č dublikuje slice_changed
        self.w.space_changed.connect(self.plot)
        self.w.slice_changed.connect(self.plot)
        self.w.redraw_called.connect(self.redraw)
        
    #č důležité je, že GLScatterPlotItem nemá .redraw()
    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self)
        
    def plot(self):
        nsim = self.w.slider.value()
        
        pos = getattr(self.w.sample_box, self.w.space)[:nsim][:,:3]
        color = np.empty((nsim, 4))
        failsi = self.w.sample_box.failsi[:nsim]
        color[failsi] = (1, 0, 0, 1)
        color[~failsi] = (0, 1, 0, 1)

        self.setData(pos=pos, color=color, size=self.w.px_size*1.5)





class UnitCube(gl.GLBoxItem):
    def __init__(self, w):
        super().__init__()
        self.w = w
        self.w.space_changed.connect(self.plot)
        self.w.redraw_called.connect(self.redraw)

        self.item = QtWidgets.QListWidgetItem('Unit cube')
        self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.item.setCheckState(QtCore.Qt.Checked)
        self.w.list_view.addItem(self.item)
        self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self)

    def plot(self):
        if self.item.checkState() and (self.w.space in ('P', 'U')):
            self.setSize(1, 1, 1)
            self.show()
        elif self.item.checkState() and (self.w.space in ('aP', 'aU')):
            x, y, z, *__ = (*self.w.sample_box.alpha,)
            self.setSize(x, y, z)
            self.show()
        else:
            self.hide()


class LastShot(gl.GLScatterPlotItem):
    def __init__(self, w):
        super().__init__()
        self.w = w
        self.w.box_runned.connect(self.plot)
        self.w.space_changed.connect(self.plot)
        self.w.redraw_called.connect(self.redraw)

        self.item = QtWidgets.QListWidgetItem('Last shot')
        self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.item.setCheckState(QtCore.Qt.Checked)
        self.w.list_view.addItem(self.item)
        self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self)

    def plot(self):
        if self.item.checkState() and (self.w.last_shot is not None):
            pos = getattr(self.w.last_shot, self.w.space)[:,:3]
            self.setData(pos=pos, color=(0, 0, 1, .7), size=self.w.px_size*2)
            self.show()
        else:
            self.hide()



#č Nekreslí to co má (vidím pouze jednu čáru)

##class Axes(gl.GLAxisItem):
##    def __init__(self, w):
##        super().__init__()
##        self.w = w
##        self.w.redraw_called.connect(self.redraw)
##
##        self.item = QtWidgets.QListWidgetItem('Axes')
##        self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
##        self.item.setCheckState(QtCore.Qt.Checked)
##        self.w.list_view.addItem(self.item)
##        self.w.list_view.itemChanged.connect(self.plot)
##        
##
##    def redraw(self):
##        self.plot()
##        self.w.central_widget.addItem(self)
##
##    def plot(self):
##        if self.item.checkState():
##            self.show()
##        else:
##            self.hide()


class Axes:
    def __init__(self, w):
        pos = np.array([[0,0,0],[1,0,0], [0,0,0],[0,1,0], [0,0,0],[0,0,1]])
        color = np.array([[255, 0, 0, 255],[0, 0, 0, 0], [0, 255, 0, 255],[0, 0, 0, 0], [0, 0, 255, 255],[0, 0, 0, 0]])
        self.axes = gl.GLLinePlotItem(pos=pos, color=color, width=2, mode='lines')
        
        self.w = w
        self.w.redraw_called.connect(self.redraw)

        self.item = QtWidgets.QListWidgetItem('Axes')
        self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.item.setCheckState(QtCore.Qt.Checked)
        self.w.list_view.addItem(self.item)
        self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self.axes)

    def plot(self):
        if self.item.checkState():
            self.axes.show()
        else:
            self.axes.hide()



class Grid:
    def __init__(self, w):
        self.gx = gl.GLGridItem()
        self.gx.setSize(10, 10, 1)
        self.gx.rotate(90, 0, 1, 0)
        #gx.translate(-5, 0, 5)
        
        self.gy = gl.GLGridItem()
        self.gy.setSize(10, 10, 1)
        self.gy.rotate(90, 1, 0, 0)
        #gy.translate(0, -10, 10)
        
        self.gz = gl.GLGridItem()
        self.gz.setSize(10, 10, 1)
        #self.gz.translate(0, 0, 0)
        
        self.w = w
        self.w.space_changed.connect(self.plot)
        self.w.redraw_called.connect(self.redraw)

        self.item = QtWidgets.QListWidgetItem('Grid')
        self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
        self.item.setCheckState(QtCore.Qt.Unchecked)
        self.w.list_view.addItem(self.item)
        self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self.gx)
        self.w.central_widget.addItem(self.gy)
        self.w.central_widget.addItem(self.gz)

    def plot(self):
        if self.item.checkState():# and (self.w.space in ( 'Rn', 'GK', 'G')):
            self.gx.show()
            self.gy.show()
            self.gz.show()
        else:
            self.gx.hide()
            self.gy.hide()
            self.gz.hide()



#č mění se význám, semantika třídy
#č teď to není nástroj pro kontrolu skříňky, ale
#č ale pro zobrazování hezkých obrázků
#č teď se mi to víc hodí do krámu
class ConvexHull(gl.GLMeshItem):
    def __init__(self, w):
        ## Mesh item will automatically compute face normals.
        super().__init__(smooth=True, shader='edgeHilight') # edgeHilight shaded
        #self.setGLOptions('opaque')
        self.nsim = -1
        self.space = ''
        
        self.w = w
        #č já asi nechcu ani snažit sa něčo zobraziť
        #č z více dimenzí
        if self.w.sample_box.nvar == 3:
            self.w.slice_changed.connect(self.plot)
            #č slice_changed postačí, zahrnuje i box_runned 
            #self.w.box_runned.connect(self.plot)
            self.w.space_changed.connect(self.plot)
            self.w.redraw_called.connect(self.redraw)

            self.failure_item = QtWidgets.QListWidgetItem('ConvexHull Failure')
            self.failure_item.setFlags(self.failure_item.flags() | QtCore.Qt.ItemIsUserCheckable)
            self.failure_item.setCheckState(QtCore.Qt.Checked)
            self.w.list_view.addItem(self.failure_item)

            self.success_item = QtWidgets.QListWidgetItem('ConvexHull Success')
            self.success_item.setFlags(self.success_item.flags() | QtCore.Qt.ItemIsUserCheckable)
            self.success_item.setCheckState(QtCore.Qt.Checked)
            self.w.list_view.addItem(self.success_item)

            
            self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self)

    def recalculate(self):
        nsim = self.w.slider.value()
        sample_box = self.w.sample_box[:nsim]

        self.space = self.w.space
        points = getattr(sample_box, self.w.space)
        qhull = spatial.ConvexHull(points, incremental=False)
        
        
        self.points = points
        self.simplices = qhull.simplices
        self.events = six.get_events(sample_box, self.simplices)
        
        nodes_colors = np.empty((nsim, 4))
        nodes_colors[sample_box.failsi] = np.array([253, 93, 97, 0])/255 
        nodes_colors[~sample_box.failsi] = np.array([67, 255, 81, 0])/255
        
        self.vertex_colors = nodes_colors

        # marker
        self.nsim = nsim

#č kus kódu jako historická památka
#č doposud si pamatuji, jak jsem tu kvůli vlástní chybě bojoval s ConvexHull'em
#č pak ConvexHull byl reabilitován
##        if (self.w.sample_box.nsim != len(self.w.sample_box.convex_hull.points)):
##            #č ConvexHull nemichá vzorky a nedělá ďupy - byla to chyba v Blackboxu
##            print(self.__class__.__name__ + \
##                  ": convex hull points mismatch. Switching to the failsafe code.")
##            sampled_plan_tri = getattr(self.w.sample_box, self.w.sample_box.tri_space)
##            tree = spatial.cKDTree(sampled_plan_tri)
##            self.points = self.w.sample_box.convex_hull.points
##            dd, ii = tree.query(self.points, k=1, p=2)
##
##            self.simplices = self.w.sample_box.convex_hull.simplices
##            box_facets = ii[self.simplices.reshape(-1)].reshape(-1, 3)
##            self.events = self.w.sample_box.get_events(box_facets)
##
##
##            nodes_colors = np.empty((self.w.sample_box.nsim, 4))
##            nodes_colors[self.w.sample_box.failsi] = np.array([253, 93, 97, 0])/255 
##            nodes_colors[~self.w.sample_box.failsi] = np.array([67, 255, 81, 0])/255
##            
##
##            self.vertex_colors = nodes_colors[ii]



    def plot(self):
        if (self.failure_item.checkState() + self.success_item.checkState())\
                               and (self.w.slider.value() > 3):
            try:
                if (self.nsim != self.w.slider.value()) or (self.space != self.w.space):
                    self.recalculate()
                    
                #face_colors = np.empty((len(self.simplices), 4), dtype=np.int16)
                # sorry for that
                if self.failure_item.checkState():
                    mask_1 = self.events==1
                else:
                    mask_1 = self.events==100500

                if self.success_item.checkState():
                    mask_0 = self.events==0
                else:
                    mask_0 = self.events==100500

                #mask_2 = self.events!=2

                mask = np.any([mask_0, mask_1], axis=0)

                
                self.setMeshData(vertexes=self.points, faces=self.simplices[mask], vertexColors=self.vertex_colors, \
                                 drawEdges=True, edgeColor=(1, 1, 0, 1))
                
                self.show()
            except BaseException as e:
                msg = "nepovedlo se nám spočítat konvexní obálku "
                error_msg = self.__class__.__name__ + ": " + msg + repr(e)
                print(error_msg)
        else:
            self.hide()
        
                
        
#č mění se význám, semantika třídy
#č teď to není nástroj pro kontrolu skříňky, ale
#č ale pro zobrazování hezkých obrázků
#č teď se mi to víc hodí do krámu
class Facets(gl.GLMeshItem):
    def __init__(self, w):
        ## Mesh item will automatically compute face normals.
        super().__init__(smooth=True, shader='shaded') # edgeHilight shaded
        #self.setGLOptions('opaque')
        self.nsim = -1
        self.space = ''
        
        self.w = w
        #č já asi nechcu ani snažit sa něčo zobraziť
        #č z více dimenzí
        if self.w.sample_box.nvar == 3: 
            self.w.slice_changed.connect(self.plot)
            #č slice_changed postačí, zahrnuje i box_runned 
            #self.w.box_runned.connect(self.plot)
            self.w.space_changed.connect(self.plot)
            self.w.redraw_called.connect(self.redraw)

            self.failure_item = QtWidgets.QListWidgetItem('Failure facets')
            self.failure_item.setFlags(self.failure_item.flags() | QtCore.Qt.ItemIsUserCheckable)
            self.failure_item.setCheckState(QtCore.Qt.Checked)
            self.w.list_view.addItem(self.failure_item)

            self.success_item = QtWidgets.QListWidgetItem('Success facets')
            self.success_item.setFlags(self.success_item.flags() | QtCore.Qt.ItemIsUserCheckable)
            self.success_item.setCheckState(QtCore.Qt.Checked)
            self.w.list_view.addItem(self.success_item)

            
            self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self)

    def recalculate(self):
        nsim = self.w.slider.value()
        sample_box = self.w.sample_box[:nsim]
        
        self.space = self.w.space
        points = getattr(sample_box, self.w.space)
        tri = spatial.Delaunay(points, incremental=False)
        
        self.points = points

        self.simplices = tri.simplices
        self.s_events = six.get_events(sample_box, self.simplices)

        self.mixed_simplices = self.simplices[self.s_events==2]

        #č ježíšmaria, co se tu děje?
        self.facets = np.vstack((self.mixed_simplices[:,:3], self.mixed_simplices[:,1:],\
                                 self.mixed_simplices[:,[0,2,3]], self.mixed_simplices[:,[0,1,3]]))

        # f_events
        self.events = six.get_events(sample_box, self.facets)
                    

        self.nodes_colors = np.empty((nsim, 4))
        self.nodes_colors[sample_box.failsi] = np.array([203, 83, 87, 0])/255 
        self.nodes_colors[~sample_box.failsi] = np.array([57, 205, 71, 0])/255
        
        
        # marker
        self.nsim = nsim


    def plot(self):
        if (self.failure_item.checkState() + self.success_item.checkState())\
                               and (self.w.slider.value() > 3):
            
            try:
                if (self.nsim != self.w.slider.value()) or (self.space != self.w.space):
                    self.recalculate()
                
                #face_colors = np.empty((len(self.simplices), 4), dtype=np.int16)
                # sorry for that
                if self.failure_item.checkState():
                    mask_1 = self.events==1
                else:
                    mask_1 = self.events==100500

                if self.success_item.checkState():
                    mask_0 = self.events==0
                else:
                    mask_0 = self.events==100500

                #mask_2 = self.events!=2

                mask = np.any([mask_0, mask_1], axis=0)

                self.setMeshData(vertexes=self.points, faces=self.facets[mask], vertexColors=self.nodes_colors, \
                                 drawEdges=True, edgeColor=(1, 1, 0, 1))
                
                self.show()
            except BaseException as e:
                msg = "nepovedlo se nám spočítat stěny "
                error_msg = self.__class__.__name__ + ": " + msg + repr(e)
                print(error_msg)
        else:
            self.hide()   

        
        

#č mění se význám, semantika třídy
#č teď to není nástroj pro kontrolu skříňky, ale
#č ale pro zobrazování hezkých obrázků
#č teď se mi to víc hodí do krámu
class Wireframe(gl.GLMeshItem):
    def __init__(self, w):
        super().__init__(smooth=False, drawFaces=False, drawEdges=True) 
        self.nsim = -1
        self.space = ''
        
        self.w = w
        #č já asi nechcu ani snažit sa něčo zobraziť
        #č z více dimenzí
        if self.w.sample_box.nvar == 3: 
            self.w.slice_changed.connect(self.plot)
            #č slice_changed postačí, zahrnuje i box_runned 
            #self.w.box_runned.connect(self.plot)
            self.w.space_changed.connect(self.plot)
            self.w.redraw_called.connect(self.redraw)

            self.item = QtWidgets.QListWidgetItem('Wireframe')
            self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
            self.item.setCheckState(QtCore.Qt.Checked)
            self.w.list_view.addItem(self.item)

            
            self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self)

    def recalculate(self):
        nsim = self.w.slider.value()
        sample_box = self.w.sample_box[:nsim]
        
        self.space = self.w.space
        points = getattr(sample_box, self.w.space)
        tri = spatial.Delaunay(points, incremental=False)
        
        self.points = points
        self.simplices = tri.simplices


        self.facets = np.vstack((self.simplices[:,:3], self.simplices[:,1:],\
                                 self.simplices[:,[0,2,3]], self.simplices[:,[0,1,3]]))



        self.setMeshData(vertexes=self.points, faces=self.facets, drawEdges=True, edgeColor=(0.5, 0.5, 0.5, 1))
        
        # marker
        self.nsim = nsim


    def plot(self):
        if self.item.checkState() and (self.w.slider.value() > 3):
            
            try:
                if (self.nsim != self.w.slider.value()) or (self.space != self.w.space):
                    self.recalculate()
                self.show()
            except BaseException as e:
                msg = "nepovedlo se nám spočítat stěny "
                error_msg = self.__class__.__name__ + ": " + msg + repr(e)
                print(error_msg)
        else:
            self.hide()
        
        


#☺ čota mne ne chočetsa eto dělať
            
##class FlexFrame(GLLinePlotItem):
##    def __init__(self, w):
##        super().__init__(smooth=False, drawFaces=False, drawEdges=True) 
##        self.nsim = -1
##        
##        pos = np.array([[0,0,0],[1,0,0], [0,0,0],[0,1,0], [0,0,0],[0,0,1]])
##        color = np.array([[255, 0, 0, 255],[0, 0, 0, 0], [0, 255, 0, 255],[0, 0, 0, 0], [0, 0, 255, 255],[0, 0, 0, 0]])
##        self.axes = gl.GLLinePlotItem(pos=pos, color=color, width=2, mode='lines')
##        
##        self.w = w
##        self.w.slice_changed.connect(self.plot)
##        self.w.space_changed.connect(self.plot)
##        self.w.redraw_called.connect(self.redraw)
##
##        self.item = QtWidgets.QListWidgetItem('FlexFrame')
##        self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
##        self.item.setCheckState(QtCore.Qt.Checked)
##        self.w.list_view.addItem(self.item)
##        self.w.list_view.itemChanged.connect(self.plot)
##
##
##    def recalculate(self):
##        self.points = self.w.sample_box.tri.points
##
##        self.simplices = self.w.sample_box.tri.simplices
##
##
##        self.facets = np.vstack((self.simplices[:,:3], self.simplices[:,1:],\
##                                 self.simplices[:,[0,2,3]], self.simplices[:,[0,1,3]]))
##
##
##
##        self.setMeshData(vertexes=self.points, faces=self.facets, drawEdges=True, edgeColor=(0.5, 0.5, 0.5, 1))
##
##        
##        
##        # take coordinates in the space, where triangulation has been performed
##        sampled_plan_tri = getattr(self.sample_box, self.sample_box.tri_space)
##        ns = 100
##        
##        if len(self.triangulation) < len(self.sample_box.tri.simplices):
##            x = y = ()
##            gap = len(self.sample_box.tri.simplices) - len(self.triangulation)
##            for __ in range(gap):
##                self.triangulation.append(self.plotWidget.plot(x, y, pen=0.7))
##        
##        for triangle, plot_item in zip(self.sample_box.tri.simplices, self.triangulation):
##            x_tri = np.linspace(sampled_plan_tri[triangle[0],0], sampled_plan_tri[triangle[1],0], ns, endpoint=False)
##            y_tri = np.linspace(sampled_plan_tri[triangle[0],1], sampled_plan_tri[triangle[1],1], ns, endpoint=False)
##            x_tri = np.append(x_tri, np.linspace(sampled_plan_tri[triangle[1],0], sampled_plan_tri[triangle[2],0], ns, endpoint=False))
##            y_tri = np.append(y_tri, np.linspace(sampled_plan_tri[triangle[1],1], sampled_plan_tri[triangle[2],1], ns, endpoint=False))
##            x_tri = np.append(x_tri, np.linspace(sampled_plan_tri[triangle[2],0], sampled_plan_tri[triangle[0],0], ns, endpoint=True))
##            y_tri = np.append(y_tri, np.linspace(sampled_plan_tri[triangle[2],1], sampled_plan_tri[triangle[0],1], ns, endpoint=True))
##            
##            tri_bound_tri = np.array((x_tri, y_tri)).T
##            # vytvořme sample
##            tri_bound = self.sample_box.sampled_plan.new_sample(tri_bound_tri, space=self.sample_box.tri_space)
##            
##            # Update the data
##            #print("Суредасько", tri_bound)
##            plot_item.setData(*getattr(tri_bound, self.space).T)
##                    
##
##
##
##        
##        # marker
##        self.nsim = self.w.sample_box.nsim
##
##
##    def redraw(self):
##        self.plot()
##        self.w.central_widget.addItem(self.axes)
##
##    def plot(self):
##        if self.item.checkState():
##            try:
##                self.recalculate()
##            except BaseException as e:
##                msg = "error during triangulation drawing"
##                error_msg = self.__class__.__name__ + ": " + msg + repr(e)
##                print(error_msg)
##                
##            self.axes.show()
##        else:
##            self.axes.hide()




class Density:
    def __init__(self, w):
        self.w = w
        #č já asi nechcu ani snažit sa něčo zobraziť
        #č z více dimenzí
        if self.w.sample_box.nvar == 3: 
            self.w.space_changed.connect(self.plot)
            self.w.redraw_called.connect(self.redraw)

            
            self.v = gl.GLVolumeItem(self.recalculate())
            self.space = self.w.space
            self.v.scale(0.1, .10, .10)
            self.v.translate(-5,-5,-5)

            self.item = QtWidgets.QListWidgetItem('Density')
            self.item.setFlags(self.item.flags() | QtCore.Qt.ItemIsUserCheckable)
            self.item.setCheckState(QtCore.Qt.Unchecked)
            self.w.list_view.addItem(self.item)

            self.w.list_view.itemChanged.connect(self.plot)
        

    def redraw(self):
        self.plot()
        self.w.central_widget.addItem(self.v)

    def recalculate(self):
        #č strašně drahý, nechce se mi to pořádně udělat
        #data = np.fromfunction(lambda i,j,k:self.w.sample_box.pdf((np.array([[i,j,k]])-[50, 50, 50])/10, self.w.space), (100,100,100))

        # ničemu nerozumím
        grid = (np.mgrid[0:100,0:100, 0:100].T.reshape(-1, 3) - [50, 50, 50])/10
        shape = (100, 100, 100)
        sample = np.array((grid[:, 2], grid[:, 1], grid[:, 0])).T
        data = self.w.sample_box.sample_pdf(sample, self.w.space)
        data = data.reshape((100, 100, 100))
        
        d2 = np.empty(data.shape + (4,), dtype=np.ubyte)
        d2[..., 0] = np.full(shape, 0)
        d2[..., 1] = np.full(shape, 200)
        d2[..., 2] = np.full(shape, 200)
        d2[..., 3] = data * (255./data.max())

        return d2
        
        


    def plot(self):
        if self.item.checkState():
            if self.w.space != self.space:
                self.v.setData(self.recalculate())
                self.space = self.w.space
            self.v.show()
        else:
            self.v.hide()   







"""
=============
Эскерон виӝет 
Widgety odhadů
Estimation widgets
===================
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
"""

        
        
class SimplexEstimationWidget(qt_plot.FastSimplexEstimationWidget):
        
        
    def callback(self, sx=None, nodes=None, cell_stats=None, simplex=None, *args, **kwargs):
        # stm trianguľaciju pokažde provadí znovu, proto skoro nemá cenu drbat se s její znovupoužitím
        if (simplex is not None) and (simplex.nvar==3):
            ns = 30
            
            
            simplex_tri = getattr(simplex, sx.tri_space)
            x_tri_1 = np.linspace(simplex_tri[0,0], simplex_tri[1,0], ns, endpoint=False)
            y_tri_1 = np.linspace(simplex_tri[0,1], simplex_tri[1,1], ns, endpoint=False)
            z_tri_1 = np.linspace(simplex_tri[0,2], simplex_tri[1,2], ns, endpoint=False)
            
            x_tri_2 = np.linspace(simplex_tri[1,0], simplex_tri[2,0], ns, endpoint=False)
            y_tri_2 = np.linspace(simplex_tri[1,1], simplex_tri[2,1], ns, endpoint=False)
            z_tri_2 = np.linspace(simplex_tri[1,2], simplex_tri[2,2], ns, endpoint=False)
            
            x_tri_3 = np.linspace(simplex_tri[2,0], simplex_tri[3,0], ns, endpoint=True)
            y_tri_3 = np.linspace(simplex_tri[2,1], simplex_tri[3,1], ns, endpoint=True)
            z_tri_3 = np.linspace(simplex_tri[2,2], simplex_tri[3,2], ns, endpoint=True)
                
            
            tri_bound_tri = np.concatenate(((x_tri_1, y_tri_1, z_tri_1),\
                                            (x_tri_2, y_tri_2, z_tri_2),\
                                            (x_tri_3, y_tri_3, z_tri_3)), axis=1).T
            # vytvořme sample
            tri_bound = self.sb_item.sample_box.sampled_plan.new_sample(tri_bound_tri,\
                                                                        space=sx.tri_space)
            
            # draw 
            pos = getattr(tri_bound, self.sb_item.space)
            plot_item = gl.GLLinePlotItem(pos=pos)
            self.sb_item.central_widget.addItem(plot_item)
            
            # uložíme data
            self.triangulation.append((tri_bound, plot_item))

            #
            #čs a eště raz
            #
            simplex_tri = getattr(simplex, sx.tri_space)
            x_tri_1 = np.linspace(simplex_tri[1,0], simplex_tri[3,0], ns, endpoint=False)
            y_tri_1 = np.linspace(simplex_tri[1,1], simplex_tri[3,1], ns, endpoint=False)
            z_tri_1 = np.linspace(simplex_tri[1,2], simplex_tri[3,2], ns, endpoint=False)
            
            x_tri_2 = np.linspace(simplex_tri[3,0], simplex_tri[0,0], ns, endpoint=False)
            y_tri_2 = np.linspace(simplex_tri[3,1], simplex_tri[0,1], ns, endpoint=False)
            z_tri_2 = np.linspace(simplex_tri[3,2], simplex_tri[0,2], ns, endpoint=False)
            
            x_tri_3 = np.linspace(simplex_tri[0,0], simplex_tri[2,0], ns, endpoint=True)
            y_tri_3 = np.linspace(simplex_tri[0,1], simplex_tri[2,1], ns, endpoint=True)
            z_tri_3 = np.linspace(simplex_tri[0,2], simplex_tri[2,2], ns, endpoint=True)
                
            tri_bound_tri = np.concatenate(((x_tri_1, y_tri_1, z_tri_1),\
                                            (x_tri_2, y_tri_2, z_tri_2),\
                                            (x_tri_3, y_tri_3, z_tri_3)), axis=1).T
            # vytvořme sample
            tri_bound = self.sb_item.sample_box.sampled_plan.new_sample(tri_bound_tri,\
                                                                        space=sx.tri_space)
            
            # draw 
            pos = getattr(tri_bound, self.sb_item.space)
            plot_item = gl.GLLinePlotItem(pos=pos)
            self.sb_item.central_widget.addItem(plot_item)
            
            # uložíme data
            self.triangulation.append((tri_bound, plot_item))



            
        
        
        #    
        # tečičky
        #

        event = cell_stats['event']
        cell_probability = cell_stats['cell_probability']
        if event in self.max_simplices:
            cm = self.param.getValues()[event][0] #č očekávám tam kolor mapu
            
            #č chcu ještě na konci prekreslit s různejma barvičkama, podle obsahu pravděpodobnosti
            # zkontrolujeme probability
            if self.max_simplices[event] < cell_probability:
                self.max_simplices[event] = cell_probability
                # a hned všecko dotyčné přebarvíme podle obsahu pravděpodobnosti
                for self_simplex in self.simplices:
                    if event == self_simplex[2]['event']:
                        # zde cell_probability se rovná self.p_cell_max[event]
                        # ale bacha! Nesplet se jinde!
                        #č nechť zůstane starý nazev
                        blue_intensity = self_simplex[2]['cell_probability'] / cell_probability
                        color = cm.mapToFloat(blue_intensity)
                        #č tam prostě MUSÍ být tuple
                        self_simplex[1].setData(color=tuple(color))

            blue_intensity = cell_probability / self.max_simplices[event]
            color = cm.mapToFloat(blue_intensity)
        else: # outside
            color = (0.6, 0.6, 0.6, 1)
        
        
        # draw tečičky
        #
        pos = getattr(nodes, self.sb_item.space)[:,:3]
        
        #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
        size = self.param.getValues()['node (pixel) size'][0]
        #č tam prostě MUSÍ být tuple
        plot_item = gl.GLScatterPlotItem(pos=pos, size=size, color=tuple(color))
        self.sb_item.central_widget.addItem(plot_item)
        
        # uložíme data
        self.simplices.append((nodes, plot_item, cell_stats))
        
        
        
        # keep the GUI responsive :)
        self.sb_item.app.processEvents()



        
            
    def recolor(self):
        with pg.BusyCursor():
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()
            #č nejdřív triangulace
            for tri_bound, plot_item in self.triangulation:
                plot_item.show()

            #č teď tečičky    
            for nodes, plot_item, cell_stats in self.simplices:
                event = cell_stats['event']
                if event in self.max_simplices:
                    cell_probability = cell_stats['cell_probability']
                    cm = self.param.getValues()[event][0] #č očekávám tam kolor mapu
                    blue_intensity = cell_probability / self.max_simplices[event]
                    color = cm.mapToFloat(blue_intensity)
                else: # outside
                    color = (0.6, 0.6, 0.6, 1)
                #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
                size = self.param.getValues()['node (pixel) size'][0]
                #č tam prostě MUSÍ být tuple
                plot_item.setData(size=size, color=tuple(color))
                plot_item.show()
        

    def on_space_changed(self, *args, **kwargs):
        # nejdřív triangulace
        for tri_bound, plot_item in self.triangulation:
            pos = getattr(tri_bound, self.sb_item.space)
            plot_item.setData(pos=pos)
            plot_item.show()

        # teď tečičky
        for nodes, plot_item, cell_stats in self.simplices:
            pos = getattr(nodes, self.sb_item.space)[:,:3]
            plot_item.setData(pos=pos)
            plot_item.show()
            





class VoronoiEstimationWidget(qt_plot.VoronoiEstimationWidget):

        

    def on_space_changed(self, *args, **kwargs):
        # teď tečičky
        for nodes, plot_item, cell_stats in self.cells:
            pos = getattr(nodes, self.sb_item.space)[:,:3]
            plot_item.setData(pos=pos)
        
    
        
        
            
        
    def node_pf_coloring(self, estimation=None, nodes=None, cell_stats=None, out_nodes=None, *args, **kwargs):
        """
        if nodes and cell_stats  provided we will add them to self.cells
        otherwise function redraw items in self.cells
        """
        
        symbol_size = self.param.getValues()['node (pixel) size'][0]
        if nodes is None:
            for nodes, plot_item, cell_stats in self.cells:
                colors = self.node_pf_colors(nodes, cell_stats)
                plot_item.setData(color=colors, size=symbol_size)
                plot_item.show()
                
        
        # máme nodes, tj. jedeme poprvé        
        else:
            colors = self.node_pf_colors(nodes, cell_stats)
            pos = getattr(nodes, self.sb_item.space)[:,:3]
            plot_item = gl.GLScatterPlotItem(pos=pos, size=symbol_size, color=colors)
            self.sb_item.central_widget.addItem(plot_item)
            
            # uložíme data
            self.cells.append((nodes, plot_item, cell_stats))
            
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()
            
            
            
            
    def node_pf_colors(self, nodes, cell_stats):
        
            
        # zas, нет ножек - нет мультиков
        # node_pf_estimations nemusejí bejt
        try:
            # zkusmě pro jednoduchost 
            # čírou RGB hračku
            npf = nodes.node_pf_estimations
            cm = pg.colormap.ColorMap((0,1), [(0, 255, 0, 255), (255, 0, 0, 255)])
            return cm.mapToFloat(npf)
            
        except BaseException as e:
            msg = "node_pf_coloring has problems "
            error_msg = self.__class__.__name__ + ": " + msg + repr(e)
            print(error_msg)
            # simple coloring
            event = cell_stats['event']
            return self.get_color(event)
        

        
        
    
    def simple_coloring(self, nodes=None, cell_stats=None, *args, **kwargs):
        """
        if nodes and cell_stats  provided we will add them to self.cells
        otherwise function redraw items in self.cells
        """
        symbol_size = self.param.getValues()['node (pixel) size'][0]
        
        if nodes is None:
            for nodes, plot_item, cell_stats in self.cells:
                event = cell_stats['event']
                color = self.get_color(event)
                #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
                plot_item.setData(color=color, size=symbol_size)
                plot_item.show()
        
        # máme nodes, tj. jedeme poprvé        
        else:
            # draw tečičky
            #
            pos = getattr(nodes, self.sb_item.space)[:,:3]
            
            event = cell_stats['event']
            color = self.get_color(event)
            #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
            plot_item = gl.GLScatterPlotItem(pos=pos, size=symbol_size, color=color)
            self.sb_item.central_widget.addItem(plot_item)
            
            # uložíme data
            self.cells.append((nodes, plot_item, cell_stats))
            
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()
    
        
            
    def cell_probability_coloring(self, nodes=None, cell_stats=None, *args, **kwargs):
        """
        if nodes and cell_stats  provided we will add them to self.cells
        otherwise function redraw items in self.cells
        """
        symbol_size = self.param.getValues()['node (pixel) size'][0]
        
        if nodes is None:
            for nodes, plot_item, cell_stats in self.cells:
                event = cell_stats['event']
                cell_probability = cell_stats['cell_probability']
                if self.p_cell_max[event] < cell_probability:
                    self.p_cell_max[event] = cell_probability
                
            # přebarvíme tečičky podle obsahu pravděpodobnosti
            for nodes, plot_item, cell_stats in self.cells:
                # draw 
                #pos = getattr(nodes, self.sb_item.space)[:,:3]
                #x, y = (*getattr(nodes, self.sb_item.space).T,)
                
                event = cell_stats['event']
                cell_probability = cell_stats['cell_probability']
                # bez modrého - maximální intenzita
                blue_intensity = 1 - cell_probability / self.p_cell_max[event]
                color = self.get_color(event, blue_intensity)
                #symbolSize = np.sqrt(nodes.w / min(nodes.w)) # not bad
                plot_item.setData(color=color, size=symbol_size)
                plot_item.show()
        
        # máme nodes, tj. jedeme poprvé        
        else:
            event = cell_stats['event']
            cell_probability = cell_stats['cell_probability']
            # zkontrolujeme probability
            if self.p_cell_max[event] < cell_probability:
                self.p_cell_max[event] = cell_probability
                # a hned všecko dotyčné přebarvíme podle obsahu pravděpodobnosti
                for cell in self.cells:
                    if event == cell[2]['event']:
                        # bez modrého - maximální intenzita
                        # zde cell_probability se rovná self.p_cell_max[event]
                        # ale bacha! Nesplet se jinde!
                        blue_intensity = 1 - cell[2]['cell_probability'] / cell_probability
                        color = self.get_color(event, blue_intensity)
                        # bacha, potřebuji prvek vložit zpätky
                        cell[1].setData(color=color)
            
            # bez modrého - maximální intenzita
            blue_intensity = 1 - cell_probability / self.p_cell_max[event]
            color = self.get_color(event, blue_intensity)
            
            
            # draw tečičky
            #
            pos = getattr(nodes, self.sb_item.space)[:,:3]
            plot_item = gl.GLScatterPlotItem(pos=pos, size=symbol_size, color=color)
            self.sb_item.central_widget.addItem(plot_item)
            
            # uložíme data
            self.cells.append((nodes, plot_item, cell_stats))
            
            # keep the GUI responsive :)
            self.sb_item.app.processEvents()


    def get_color(self, event, blue_intensity=None):
        """
        get color for 'simple_coloring' or 'cell_probability_coloring'
        """
        # generally
        if event == 'success':
            color = [167, 255, 181]# xkcd:light seafoam green #a7ffb5
        elif event == 'failure':
            color = [253, 193, 197] # xkcd: pale rose (#fdc1c5)
        # já vím, že Voronoi nemá 'mix', эн но юа
        else:# 'mix'
            color = [255, 243, 154] # let's try xkcd: dark cream (#fff39a)
            
        if blue_intensity is not None:
            # let's play with blue a little bit
            # chcu mít korektní výstup
            # aspoň v něčem chcu být jistý
            if blue_intensity > 1:
                color[2] = 255
            elif blue_intensity < 0:
                color[2] = 0
            else:
                # pyqtgraph žere barvy i s čarkou
                # ale my je mu davat nebudeme
                color[2] = int(blue_intensity*255) 
            
        return (*(np.array(color)/255), 1)


"""
===========
♥ Чыры-пыры 
č Jiné
E Miscellaneous
===============
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
"""

        
        
class CandidatesWidget(qt_plot.CandidatesWidget):


    def run_stm(self):
        #č indikace
        #self.setDisabled(True)
        with pg.BusyCursor():
            
            color_map = self.gradient.colorMap()
            
            try:#č může se tu stat cokoliv
                #č načíst sloupce prostě z libovolného vzorku
                cb = self.sb_item.sample_box.candidates_index.values().__iter__().__next__()
                #č je třeba mít seznam aktualní
                self.attr.setItems(self.get_items())
                
                #č neplest s self.attr!
                attr = self.attr.currentText()
                
                #č kruci, nejdřív je třeba najít maxvalue, minvalue je implicitně nula
                maxvalue = -np.inf
                minvalue = np.inf
                for id, cb in self.sb_item.sample_box.candidates_index.items():
                    array = np.atleast_1d(getattr(cb, attr))
                    if len(array):
                        maxcb = np.nanmax(array)
                        mincb = np.nanmin(array)
                        if maxcb > maxvalue:
                            maxvalue = maxcb
                            maxitem = cb[np.nanargmax(array)]
                        if mincb < minvalue:
                            minvalue = mincb
                
                #č zvlášť nakreslím maximální hodnotu
                pos = getattr(maxitem, self.sb_item.space)[:,:3]
                max_item = gl.GLScatterPlotItem(pos=pos, size=self.sb_item.px_size*2, color=color_map.mapToFloat(1))
                self.sb_item.central_widget.addItem(max_item)
                self.pens.append(max_item)
                
                
                
                #č a teď jdeme!
                for id, cb in self.sb_item.sample_box.candidates_index.items():
                    array = np.atleast_1d(getattr(cb, attr))
                    if np.isnan(array).any():
                        msg = "%s candidates has nans in %s attribute"%(id, attr)
                        error_msg = self.__class__.__name__ + ": " + msg
                        print(error_msg)
                    mask = np.isfinite(array)
                    values = array[mask]
                    norm_values = (values - minvalue) / (maxvalue - minvalue)
                    
                    pos = getattr(cb, self.sb_item.space)[mask][:,:3]
                    color = color_map.mapToFloat(norm_values)
                    pen = gl.GLScatterPlotItem(pos=pos, size=self.sb_item.px_size, color=color)
                    self.sb_item.central_widget.addItem(pen)
                    self.pens.append(pen)
                
            except BaseException as e:
                msg = ""
                error_msg = self.__class__.__name__ + ": " + msg + repr(e)
                print(error_msg)
            
            
            
        # indikace
        #self.setEnabled(True)
        

    
        

        
